# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gdk, Gio, GObject, Gtk

from iotas.editor_search_entry import EditorSearchEntry
from iotas.ui_utils import check_for_search_starting


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/render_search_header_bar.ui")
class RenderSearchHeaderBar(Adw.Bin):
    __gtype_name__ = "RenderSearchHeaderBar"
    __gsignals__ = {
        "resumed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    _entry: EditorSearchEntry = Gtk.Template.Child()
    _backward_button: Gtk.Button = Gtk.Template.Child()
    _forward_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self.__active = False

        self.__current_matches = 0
        self.__current_match = 0
        self.__current_term = ""

        self.__limit_notify_timeout = None
        self.__found_signal_handler_id = None
        self.__failed_to_find_signal_handler_id = None
        self.__term_changed_signal_handler_id = None

    def setup(self) -> None:
        """Perform initial setup."""
        self.__setup_actions()

    def check_if_starting(
        self, controller: Gtk.EventControllerKey, keyval: int, state: Gdk.ModifierType
    ) -> bool:
        """Check if starting search via type to search.

        :param Gtk.EventControllerKey controller: The key controller
        :param int keyval: The key value
        :param Gdk.ModifierType state: Any modifier state
        :return: Whether search is starting
        :rtype: bool
        """
        if not check_for_search_starting(controller, keyval, state):
            return False

        self._entry.set_text("")
        if controller.forward(self._entry.text):
            return True

        return False

    def enter(self, render_view, resuming: bool, type_to_search: bool) -> None:
        """Enter search.

        The render view is passed in each time as, depending on configuration, it may be a new
        instance each time.

        :param bool render_view: WebKit view
        :param bool resuming: Whether resuming search (via keyboard shortcut)
        :param bool type_to_search: Starting from type to search
        """
        if type_to_search or resuming:
            # Used when starting search by typing; the initial character will already be populated
            # in the GtkText.
            self._entry.text.grab_focus()
            self._entry.text.set_position(-1)
        else:
            self._entry.select_all_and_focus()

        if self.__found_signal_handler_id is None:
            self.__render_view = render_view
            self.__controller = render_view.get_find_controller()

            if not type_to_search and not resuming:
                self._entry.set_text("")

            self.__term_changed_signal_handler_id = self._entry.text.get_buffer().connect(
                "notify::text", lambda _o, _v: self.__do_search()
            )
            self.__found_signal_handler_id = self.__controller.connect(
                "found-text", self.__on_found_text
            )
            self.__failed_to_find_signal_handler_id = self.__controller.connect(
                "failed-to-find-text", lambda _c: self.__on_failed_to_find_text()
            )

            self.__current_matches = 0
            self.__current_match = 0
            self.__occurrences_count_changed()
            # If we're here via type-to-search we may already have some text
            self.__current_term = self._entry.get_text()

            # Part of the effort to delay all WebKit initialisation
            if "WebKit" not in globals():
                import gi

                gi.require_version("WebKit", "6.0")
                global WebKit
                from gi.repository import WebKit

            if resuming:
                # Blank to avoid found callback thinking we're just iterating results
                self.__current_term = ""
                self.__do_search()
            else:
                self.__resuming = False

        self.active = True

    def exit(self) -> None:
        """Exit search."""
        self.__controller.search_finish()
        self.__controller.disconnect(self.__found_signal_handler_id)
        self.__found_signal_handler_id = None
        self.__controller.disconnect(self.__failed_to_find_signal_handler_id)
        self._entry.text.get_buffer().disconnect(self.__term_changed_signal_handler_id)
        self.active = False

    def disable_actions(self) -> None:
        """Disable actions."""
        self.__action_group.lookup_action("backward").set_enabled(False)
        self.__action_group.lookup_action("forward").set_enabled(False)

    def focus(self) -> None:
        """Grab focus to entry."""
        self._entry.text.grab_focus()

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("forward")
        action.connect("activate", lambda _o, _p: self.__move_forward())
        action_group.add_action(action)
        app.set_accels_for_action("render-search.forward", ["<Control>g"])
        action.set_enabled(False)

        action = Gio.SimpleAction.new("backward")
        action.connect("activate", lambda _o, _p: self.__move_backward())
        action_group.add_action(action)
        app.set_accels_for_action("render-search.backward", ["<Shift><Control>g"])
        action.set_enabled(False)

        self.__action_group = action_group
        app.get_active_window().insert_action_group("render-search", action_group)

    def __on_found_text(self, controller, match_count: int) -> None:
        if controller.get_search_text() == self.__current_term:
            return

        self.__current_term = controller.get_search_text()
        self.__current_matches = match_count
        self.__occurrences_count_changed()
        self.__current_match = 1
        self._entry.set_occurrence_position(1)

    def __on_failed_to_find_text(self) -> None:
        self.__current_match = 0
        self.__current_matches = 0
        self.__occurrences_count_changed()

    def __do_search(self) -> None:
        self.__controller.search(
            self._entry.get_text(),
            # mypy ignore to allow for module lazy loading
            WebKit.FindOptions.CASE_INSENSITIVE | WebKit.FindOptions.WRAP_AROUND,  # type: ignore [name-defined] # noqa: E501
            2**32 - 1,
        )

    def __move_forward(self) -> None:
        """Move to next search match."""

        if not self.__active and self._entry.get_text() != "":
            self.emit("resumed")
            return

        self.__current_match += 1
        if self.__current_match > self.__current_matches:
            self.__current_match = 1
        self._entry.set_occurrence_position(self.__current_match)
        self.__controller.search_next()

    def __move_backward(self) -> None:
        """Move to previous search match."""

        if not self.__active and self._entry.get_text() != "":
            self.emit("resumed")
            return

        self.__current_match -= 1
        if self.__current_match < 1:
            self.__current_match = self.__current_matches
        self._entry.set_occurrence_position(self.__current_match)
        self.__controller.search_previous()

    def __occurrences_count_changed(self) -> None:
        self._entry.set_occurrence_count(self.__current_matches)
        search_can_move = self.__current_matches > 0
        self.__action_group.lookup_action("backward").set_enabled(search_can_move)
        self.__action_group.lookup_action("forward").set_enabled(search_can_move)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gi.repository import GLib

import datetime
import json
import logging
import os
import re
from typing import Optional
import unicodedata

from iotas.markdown_helpers import get_image_attachments_from_note_content
from iotas.note import Note


def build_default_filename(
    note: Note, out_format: str, file_extension: str, add_timestamp: bool = False
) -> str:
    """Build an export filename for the note.

    :param Note note: Note to export
    :param str out_format: Export format
    :param str file_extension: File extension
    :param bool add_timestamp: Whether to prefix timestamp
    :return: Filename
    :rtype: str
    """
    filename = sanitise_title_for_filename(note.title)
    if add_timestamp:
        ts = datetime.datetime.now()
        filename = ts.strftime("%Y-%m-%dT%H:%M:%S") + " " + filename
    if not exporting_to_dir(out_format, note):
        filename += "." + file_extension
    return filename


def exporting_to_dir(out_format: str, note: Note) -> bool:
    """Determine whether note export to format will result in a directory.

    :param str out_format: Format
    :param Note note: Note to export
    :return: Whether exporting to dir
    :rtype: bool
    """
    if out_format == "md":
        export_to_dir = len(get_image_attachments_from_note_content(note)) > 0
    elif out_format == "html":
        export_to_dir = True
    else:
        export_to_dir = False
    return export_to_dir


def build_batch_dirname(ts: Optional[datetime.datetime] = None) -> str:
    """Build a directory name for bulk export.

    :param datetime.datetime ts: Timestamp, optional
    :return: Directory name
    :rtype: str
    """
    if not ts:
        ts = datetime.datetime.now()
    date = ts.strftime("%Y-%m-%d")
    time = ts.strftime("%H:%M:%S")
    # Translators: Directory name for batch export, {0} is a date (YYYY-MM-DD),
    # {1} is a time (HH:MM:SS)
    return _("{0} at {1} batch").format(date, time)


def sanitise_title_for_filename(title: str) -> str:
    """Sanitise title for a filesystem filename.

    For synced notes the server has already done the sanitising for us. This is for
    local-only instances.

    :param str title: Input title
    :return: Sanitised title
    :rtype: str
    """
    value = unicodedata.normalize("NFKC", str(title))
    return re.sub(r"[^\w\s\.-]", "", value).strip()


def parse_export_formats(formats: str) -> Optional[dict[str, dict[str, str]]]:
    """Parse custom Pandoc export formats from configuration string.

    :param str formats: Formats JSON string
    :return: Dict with Iotas format id as key and a dict with the Pandoc format and file extension
        as value, on success
    :rtype: dict[str, dict[str, str]]
    """
    if formats.strip() == "":
        return None
    try:
        formats_json = json.loads(formats)
    except json.JSONDecodeError as e:
        logging.warning(f"Failed to load extra formats: {e.msg}")
        return None

    if type(formats_json) is not list:
        logging.warning("Failed to load extra formats: JSON not being a list")
        return None

    out = {}

    for item in formats_json:
        if type(item) is not dict:
            logging.warning("Failed to load extra format: item not being an object")
            continue
        if "pandocOutFormat" not in item:
            logging.warning("Failed to load extra format: missing pandocOutFormat")
            continue
        if "fileExtension" not in item:
            logging.warning("Failed to load extra format: missing fileExtension")
            continue

        out_format = item["pandocOutFormat"]
        extension = item["fileExtension"]
        if type(out_format) is not str or out_format.strip() == "":
            logging.warning("Failed to load extra format: pandocOutFormat is not a string")
            continue
        if type(extension) is not str or extension.strip() == "":
            logging.warning("Failed to load extra format: due to fileExtension is not a string")
            continue

        if re.match(r"^[\w_]+$", out_format) is None:
            logging.warning(
                "Failed to load extra format: pandocOutFormat contains invalid characters"
            )
            continue
        if re.match(r"^[\w]+$", extension) is None:
            logging.warning(
                "Failed to load extra format: fileExtension contains invalid characters"
            )
            continue

        logging.debug(f"Adding extra pandoc format {out_format} with extension {extension}")

        format_id = out_format.lower()
        out[format_id] = {"format": out_format, "extension": extension}

    return out


def get_container_export_dir() -> str:
    """Get export directory within the user data directory.

    "Typically in the Flatpak container.

    :rtype: str
    """
    return os.path.join(
        GLib.get_user_data_dir(),
        "iotas",
        "exports",
    )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
import unittest
import warnings

from iotas.note import Note
from iotas.export_utils import (
    build_batch_dirname,
    build_default_filename,
    exporting_to_dir,
    parse_export_formats,
    sanitise_title_for_filename,
)

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_build_default_filename(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        filename = build_default_filename(
            note, out_format="md", file_extension="md", add_timestamp=False
        )
        self.assertEqual(filename, "Test Note Title.md")

        filename = build_default_filename(
            note, out_format="odt", file_extension="odt", add_timestamp=False
        )
        self.assertEqual(filename, "Test Note Title.odt")

        filename = build_default_filename(
            note, out_format="pdf", file_extension="pdf", add_timestamp=False
        )
        self.assertEqual(filename, "Test Note Title.pdf")

        filename = build_default_filename(
            note, out_format="html", file_extension="html", add_timestamp=False
        )
        self.assertEqual(filename, "Test Note Title")

        filename = build_default_filename(
            note, out_format="md", file_extension="md", add_timestamp=True
        )
        parts = filename.split(" ")
        self.assertEqual(len(parts), 4)

        timestamp = None
        try:
            timestamp = datetime.strptime(parts[0], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            pass

        self.assertIsNotNone(timestamp)
        self.assertEqual(" ".join(parts[1:]), "Test Note Title.md")

    def test_exporting_to_dir(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        self.assertTrue(exporting_to_dir("html", note))
        self.assertFalse(exporting_to_dir("md", note))

        note.content = "![image](.attachments.1/image.png)"

        self.assertTrue(exporting_to_dir("md", note))

    def test_build_batch_dirname(self) -> None:
        now = datetime.now()
        name = build_batch_dirname(now)
        self.assertEqual(
            name, "{0} at {1} batch".format(now.strftime("%Y-%m-%d"), now.strftime("%H:%M:%S"))
        )

    def test_sanitise_title_for_filename(self) -> None:
        title = "!@#$%^&*()+word={}[]|\\?></,"
        sanitised = sanitise_title_for_filename(title)
        self.assertEqual(sanitised, "word")

    def test_parse_export_formats(self) -> None:
        self.assertIsNone(parse_export_formats(""))
        self.assertIsNone(parse_export_formats("plain"))

        json = """
        [
          {"pandocOutFormat": "docx", "fileExtension": "docx"},
          {"pandocOutFormat": "epub3", "fileExtension": "epub"}
        ]
        """
        parsed = parse_export_formats(json)
        assert parsed
        self.assertTrue(type(parsed) is dict)
        self.assertTrue("docx" in parsed)
        self.assertTrue("epub3" in parsed)
        self.assertFalse("epub" in parsed)
        self.assertEqual(len(parsed["docx"]), 2)
        self.assertEqual(parsed["epub3"]["format"], "epub3")
        self.assertEqual(parsed["epub3"]["extension"], "epub")

        invalid = """
        [
          {"pandocOutFormatx": "docx", "fileExtension": "docx"}
        ]
        """
        self.assertEqual(parse_export_formats(invalid), {})



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib, Gtk

import logging
from typing import Any, Callable, Optional

from iotas.note import Note
from iotas.pdf_exporter import PdfExporter


class WebKitPdfExporter(PdfExporter):

    # webkit type hinting avoided to allow for module lazy loading
    def __init__(self, render_view, keep_webkit_process: bool) -> None:
        super().__init__()
        self.__render_view = render_view
        self.__render_view.connect("loaded", lambda _o: self.__on_web_view_loaded())
        self.__keep_webkit_process = keep_webkit_process
        self.__finished_callback: Callable[[], None]
        self.__error_callback: Callable[[str], None]
        self.__active = False
        self.__in_error = False

    def set_callbacks(
        self, finished_callback: Callable[[], None], error_callback: Callable[[str], None]
    ) -> None:
        """Set functions to be called upon export result.

        :param Callable finished_callback: Finished callback
        :param Callable error_callback: Error callback
        """
        self.__finished_callback = finished_callback
        self.__error_callback = error_callback

    def export(self, note: Note, location: str) -> None:
        """Export PDF of note.

        :param Note note: Note to export
        :param str location: Destination location
        """
        self.__note = note
        self.__location = location
        self.__active = True
        self.__in_error = False

        # Part of the effort to delay all WebKit initialisation
        if "WebKit" not in globals():
            import gi

            gi.require_version("WebKit", "6.0")
            global WebKit
            from gi.repository import WebKit

        self.__render_view.render_retaining_scroll(self.__note, "pdf")

    def __on_web_view_loaded(self) -> None:
        if not self.__active:
            return

        printer_name = self.__get_printer_name()
        if printer_name is None:
            self.__in_error = True
            # Theoretically rare, avoiding translation for now
            self.__error_callback("Failed to locate virtual printer for export to PDF")
            return

        settings = Gtk.PrintSettings()
        settings.set(Gtk.PRINT_SETTINGS_PRINTER, printer_name)
        settings.set(Gtk.PRINT_SETTINGS_OUTPUT_FILE_FORMAT, "pdf")
        settings.set(Gtk.PRINT_SETTINGS_OUTPUT_URI, f"file://{self.__location}")

        page_setup = Gtk.PageSetup()
        page_setup.set_left_margin(54, Gtk.Unit.POINTS)
        page_setup.set_right_margin(54, Gtk.Unit.POINTS)

        operation = WebKit.PrintOperation.new(self.__render_view)  # type: ignore [name-defined]
        operation.set_print_settings(settings)
        operation.set_page_setup(page_setup)

        def finished(_print_operation):
            logging.info(f"Exported pdf to {self.__location}")
            if not self.__keep_webkit_process:
                logging.info("Terminating WebKit process as holding disabled in preference")
                self.__render_view.terminate_web_process()
            if not self.__in_error:
                self.__finished_callback()
            self.__active = False

        def failed(_print_operation, error: GLib.Error):
            logging.warning(f"Failed to export pdf to {self.__location}: {error.message}")
            if not self.__keep_webkit_process:
                logging.info("Terminating WebKit process as holding disabled in preference")
                self.__render_view.terminate_web_process()
            self.__in_error = True
            self.__error_callback(error.message)

        operation.connect("finished", finished)
        operation.connect("failed", failed)
        operation.print_()

    def __get_printer_name(self) -> Optional[str]:
        default_language = Gtk.get_default_language().to_string()
        if default_language.startswith("en"):
            return "Print to File"
        else:
            virtual_printers = []

            def check_printer(printer: Gtk.Printer, _data: Any) -> None:
                if printer.is_virtual():
                    virtual_printers.append(printer.get_name())

            Gtk.enumerate_printers(check_printer, data=None, wait=True)
            if len(virtual_printers) == 1:
                return virtual_printers[0]
            elif not virtual_printers:
                logging.warning("No virtual printers found")
                return None
            else:
                logging.warning("See more than one virtual printer, can't identify Print to File")
                return None



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, timedelta
import os
import shutil
import unittest
import warnings

from iotas.category import Category, CategorySpecialPurpose
from iotas.category_manager import CategoryManager
from iotas.database import Database
from iotas.note import Note
from iotas.note_database import NoteDatabase
import iotas.note_manager
from iotas.note_manager import NoteManager

warnings.filterwarnings("ignore", "version")

# Note: very focused testing here so far purely for issues with the filters for the note list
# sections. Coverage very low and even the filter testing is narrow.


class Test(unittest.TestCase):

    note_fake_id = 1
    now: datetime

    def test_filter_cutoff_duplicates(self) -> None:
        # Typical thresholds
        self.__reset(datetime(2025, 9, 19, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(0))
        notes.append(self.__create_dummy_note(1))
        notes.append(self.__create_dummy_note(2))
        notes.append(self.__create_dummy_note(4))
        notes.append(self.__create_dummy_note(5))
        notes.append(self.__create_dummy_note(18))
        notes.append(self.__create_dummy_note(19))
        notes.append(self.__create_dummy_note(49))
        notes.append(self.__create_dummy_note(50))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [1, 1, 2, 2, 2, 1])

        # Fishing over a year
        # start = datetime(2025, 9, 15, 14, 11, 0)
        # for ago in range(0, 366):
        #     dt = start - timedelta(days=ago)
        #     self.__reset(dt)
        #     notes = []
        #     notes.append(self.__create_dummy_note(0))
        #     notes.append(self.__create_dummy_note(1))
        #     notes.append(self.__create_dummy_note(2))
        #     notes.append(self.__create_dummy_note(4))
        #     notes.append(self.__create_dummy_note(5))
        #     notes.append(self.__create_dummy_note(18))
        #     notes.append(self.__create_dummy_note(19))
        #     notes.append(self.__create_dummy_note(49))
        #     notes.append(self.__create_dummy_note(50))
        #     notes.append(self.__create_dummy_note(80))
        #     self.__add_and_invalidate(notes)
        #     self.assertEqual(self.__get_models_sum(), 10)

        # First day of week is first day of month
        self.__reset(datetime(2025, 9, 4, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(0))
        notes.append(self.__create_dummy_note(1))
        notes.append(self.__create_dummy_note(2))
        notes.append(self.__create_dummy_note(3))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [1, 1, 2, 0, 0, 0])

        # Yesterday is first day of month
        self.__reset(datetime(2025, 8, 2, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(1))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [0, 1, 0, 0, 0, 0])

        # Yesterday is first day of month and first day of month
        self.__reset(datetime(2024, 7, 2, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(0))
        notes.append(self.__create_dummy_note(1))
        notes.append(self.__create_dummy_note(2))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [1, 1, 0, 0, 1, 0])

        # Yesterday is last week
        self.__reset(datetime(2025, 9, 15, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(1))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [0, 1, 0, 0, 0, 0])

        # Yesterday is last month
        self.__reset(datetime(2025, 8, 1, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(1))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [0, 1, 0, 0, 0, 0])

        # Yesterday is last week and last month
        self.__reset(datetime(2025, 9, 1, 14, 11, 0))
        notes = []
        notes.append(self.__create_dummy_note(1))
        self.__add_and_invalidate(notes)
        self.assertEqual(self.__get_model_totals(), [0, 1, 0, 0, 0, 0])

        self.__clean_output_dir()

    def __create_dummy_note(self, days_prior: int) -> Note:
        note = Note(new_note=True)
        note.title = note.content = f"{days_prior} days prior"
        note.last_modified = self.__create_timestamp(days_prior, self.now)
        note.id = self.note_fake_id
        self.note_fake_id += 1
        return note

    def __create_timestamp(self, days_prior: int, now: datetime) -> float:
        val = now - timedelta(days=days_prior)
        # logging.warning(f"{days_prior} days before {now} is {val}")
        return int(val.timestamp())

    def __reset(self, now: datetime) -> NoteManager:
        self.now = now
        self.__prepare_output_dir()
        storage_dir = self.__get_output_dir()
        self.db = NoteDatabase(Database(storage_dir))
        self.category_manager = CategoryManager(self.db)
        self.note_manager = NoteManager(self.db, self.category_manager)
        self.note_manager.initiate_older_notes_model()
        self.note_fake_id = 1
        iotas.note_manager.get_now = self.__get_test_now
        return self.note_manager

    def __prepare_output_dir(self) -> str:
        out_path = self.__get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)
        os.makedirs(out_path)
        return out_path

    def __clean_output_dir(self) -> None:
        out_path = self.__get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)

    def __get_output_dir(self) -> str:
        file_dir = os.path.dirname(__file__)
        return os.path.join(file_dir, os.pardir, "testing-tmp")

    def __get_model_totals(self) -> list[int]:
        # logging.warning(f"Today {self.note_manager.today_model.get_n_items()}")
        # logging.warning(f"Yesterday {self.note_manager.yesterday_model.get_n_items()}")
        # logging.warning(f"Week {self.note_manager.week_model.get_n_items()}")
        # logging.warning(f"Month {self.note_manager.month_model.get_n_items()}")
        # logging.warning(f"Last month {self.note_manager.last_month_model.get_n_items()}")
        # logging.warning(f"Older {self.note_manager.older_notes_model.get_n_items()}")
        models = [
            self.note_manager.today_model,
            self.note_manager.yesterday_model,
            self.note_manager.week_model,
            self.note_manager.month_model,
            self.note_manager.last_month_model,
            self.note_manager.older_notes_model,
        ]
        totals = []
        for model in models:
            totals.append(model.get_n_items())
        return totals

    def __get_models_sum(self) -> int:
        return sum(self.__get_model_totals())

    def __update_filters(self) -> None:
        all_category = Category("", 0)
        all_category.special_purpose = CategorySpecialPurpose.ALL
        self.note_manager.update_filters(all_category, True, False)

    def __get_test_now(self) -> datetime:
        return self.now

    def __add_and_invalidate(self, notes: list[Note]) -> None:
        self.note_manager.add_notes_to_model(notes)
        self.__update_filters()



from functools import partial

from markdown_it import MarkdownIt
from markdown_it.common.utils import arrayReplaceAt
from markdown_it.rules_core import StateCore
from markdown_it.token import Token


def img_figures_plugin(md: MarkdownIt, lazyload: bool) -> None:
    """Plugin which wraps images with plain text descriptions in figures.

    :param MarkdownIt md: Parser
    :param bool lazyload: Whether to set loading to lazy on images
    """

    def _img_figures(state: StateCore, lazyload: bool) -> None:
        for inline_token in state.tokens:
            if inline_token.type != "inline":
                continue

            tokens = inline_token.children

            assert tokens is not None
            i = len(tokens)
            while i >= 1:
                i -= 1
                assert isinstance(tokens, list)
                current_token = tokens[i]

                if current_token.type == "image":
                    if lazyload:
                        current_token.attrs["loading"] = "lazy"
                    if current_token.children:
                        nodes = []
                        level = current_token.level

                        token = Token("figure_open", "figure", 1)
                        token.level = level
                        level += 1
                        nodes.append(token)

                        nodes.append(current_token)

                        token = Token("figcaption_open", "figcaption", 0)
                        token.level = level
                        level += 1
                        token.attrs = {"aria-hidden": "true"}
                        token.content = current_token.content
                        nodes.append(token)

                        token = Token("text", "", 0)
                        token.content = current_token.content
                        token.level = level
                        nodes.append(token)

                        token = Token("figcaption_close", "figcaption", 0)
                        token.level = level
                        level -= 1
                        nodes.append(token)

                        token = Token("figure_close", "figure", -1)
                        level -= 1
                        token.level = level

                        nodes.append(token)

                        inline_token.children = tokens = arrayReplaceAt(tokens, i, nodes)

    md.core.ruler.after(
        "inline",
        "img-figures",
        partial(_img_figures, lazyload=lazyload),
    )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib, GObject

from collections import OrderedDict
import logging
from typing import Optional

from packaging.version import Version, InvalidVersion

from iotas.config_manager import ConfigManager, HeaderBarVisibility
import iotas.const as const
from iotas.note import DirtyFields
from iotas.note_database import NoteDatabase
from iotas.ui_utils import have_window_with_width, is_likely_mobile_device


class MigrationAssistant(GObject.Object):

    def __init__(self, db: NoteDatabase) -> None:
        self.__db = db
        self.__config_manager = ConfigManager.get_default()

        self.__migrations = OrderedDict()
        self.__migrations["0.1.14"] = self.__make_keeping_webkit_in_memory_default
        self.__migrations["0.9.0"] = self.__v090_preference_migrations

        self.__ui_dependent_migrations = OrderedDict()
        self.__ui_dependent_migrations["0.9.0"] = self.__set_formatting_bar_style_based_on_device

        self.__previous_version: Optional[Version] = None
        version = self.__config_manager.last_launched_version
        try:
            self.__previous_version = Version(version)
        except InvalidVersion:
            logging.warning(f"Unexpected last version {version}")

    def migrate(self) -> None:
        current_version = self.__get_current_version()
        if self.__previous_version != current_version:
            if self.__previous_version is not None:
                if current_version > self.__previous_version:
                    logging.info("Updating to v{}".format(current_version))
                    self.__run_migrations(self.__migrations)
                    self.__wait_and_run_ui_dependent_migrations()
                else:
                    logging.info(
                        "Downgrading to v{}? Not running migrations.".format(current_version)
                    )
                self.__regenerate_excerpts()
            self.__config_manager.last_launched_version = str(current_version)

    def __wait_and_run_ui_dependent_migrations(self) -> None:

        # Another location (ala FirstStart) using this pattern as a result of not yet
        # discovering the clean way to be notified when a window has obtained its initial size
        def wait_and_run():
            if not have_window_with_width():
                GLib.timeout_add(250, wait_and_run)
            else:
                self.__run_migrations(self.__ui_dependent_migrations)

        wait_and_run()

    def __run_migrations(self, migrations: OrderedDict) -> None:
        assert self.__previous_version is not None  # mypy
        for version, function in migrations.items():
            if Version(version) > self.__previous_version:
                function()

    def __get_current_version(self) -> Version:
        version = const.VERSION
        # Handle development suffixes
        if "-" in version:
            version = version.split("-")[0]
        return Version(version)

    def __regenerate_excerpts(self) -> None:
        notes = self.__db.get_all_notes(load_content=True)
        for note in notes:
            if note.regenerate_excerpt():
                changed_fields = DirtyFields()
                changed_fields.excerpt = True
                self.__db.persist_note_selective(note, changed_fields)

    def __make_keeping_webkit_in_memory_default(self) -> None:
        logging.info("Setting WebKit process to be retained in memory after first use")
        self.__config_manager.markdown_keep_webkit_process = True

    def __v090_preference_migrations(self) -> None:
        logging.info("Running v0.9.0 preference migrations")
        if not self.__config_manager.get_markdown_syntax_hightlighting_enabled():
            logging.info(
                "Syntax highlighting was previous disabled, migrating to syntax detection enabled "
                "with unstyled theme"
            )
            self.__config_manager.markdown_detect_syntax = True
            self.__config_manager.editor_theme = "iotas-unstyled"

        if self.__config_manager.get_hide_editor_headerbar():
            visibility = HeaderBarVisibility.AUTO_HIDE
        elif self.__config_manager.get_hide_editor_headerbar_when_fullscreen():
            visibility = HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY
        else:
            visibility = HeaderBarVisibility.ALWAYS_VISIBLE
        logging.info(
            f"Migrated previous header bar preferences to '{str(visibility)}' header bar visibility"
        )
        self.__config_manager.editor_header_bar_visibility = visibility

    def __set_formatting_bar_style_based_on_device(self) -> None:
        if is_likely_mobile_device():
            logging.info("Likely on mobile device, setting formatting bar to stay visible")
            visibility = HeaderBarVisibility.ALWAYS_VISIBLE
        else:
            logging.info("Likely not on mobile device, setting formatting bar to auto hide")
            visibility = HeaderBarVisibility.AUTO_HIDE
        self.__config_manager.editor_formatting_bar_visibility = visibility



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gi.repository import Adw, Gdk, Gio, GLib, Gtk

from enum import IntEnum, auto
from threading import Thread
from typing import Optional

from iotas.config_manager import ConfigManager
from iotas.nextcloud_sync_worker import LoginRequestResult
from iotas.sync_manager import SyncManager


class LoginDialogMode(IntEnum):
    NORMAL = auto()
    REAUTHENTICATE = auto()
    SECRET_SERVICE_FAILURE = auto()


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/nextcloud_login_dialog.ui")
class NextcloudLoginDialog(Adw.Dialog):
    __gtype_name__ = "NextcloudLoginDialog"

    CHECK_LOGIN_INTERVAL = 1000

    _toolbar_view: Adw.ToolbarView = Gtk.Template.Child()
    _stack: Gtk.Stack = Gtk.Template.Child()

    _intro: Adw.StatusPage = Gtk.Template.Child()
    _secret_service_failure: Adw.StatusPage = Gtk.Template.Child()
    _entry: Adw.ToastOverlay = Gtk.Template.Child()
    _self_signed_certificate_handling: Adw.StatusPage = Gtk.Template.Child()
    _loading: Adw.StatusPage = Gtk.Template.Child()
    _finished: Adw.StatusPage = Gtk.Template.Child()

    _url_continue_button: Gtk.Button = Gtk.Template.Child()
    _url_entry: Adw.EntryRow = Gtk.Template.Child()

    _instance_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, sync_manager: SyncManager, mode: LoginDialogMode) -> None:
        super().__init__()
        self.__sync_manager = sync_manager
        self.__config_manager = ConfigManager.get_default()
        self.__mode = mode
        self.__timeout_id = None
        if mode == LoginDialogMode.SECRET_SERVICE_FAILURE:
            self._stack.set_visible_child(self._secret_service_failure)
        elif mode == LoginDialogMode.REAUTHENTICATE:
            endpoint = self.__config_manager.nextcloud_endpoint
            if endpoint != "":
                self._url_entry.set_text(endpoint)
            self.__attempt_login()

        self.__sync_manager.connect("finished", self.__on_initial_sync_finished)

        self.connect("closed", lambda _o: self.__on_close())

    @Gtk.Template.Callback()
    def _on_intro_continue_clicked(self, _button: Gtk.Button) -> None:
        self._stack.set_visible_child(self._entry)
        self._url_continue_button.set_sensitive(False)
        endpoint = self.__config_manager.nextcloud_endpoint
        if endpoint != "":
            self._url_entry.set_text(endpoint)
        self._url_entry.grab_focus()
        self._url_entry.set_position(-1)

    @Gtk.Template.Callback()
    def _on_entry_continue_clicked(self, _button: Gtk.Button) -> None:
        self.__attempt_login()

    @Gtk.Template.Callback()
    def _on_url_entry_changed(self, _entry: Gtk.Editable) -> None:
        self._url_continue_button.set_sensitive(self.__url_entry_valid())

    @Gtk.Template.Callback()
    def _on_url_entry_activated(self, _entry: Gtk.Entry) -> None:
        if self.__url_entry_valid():
            self.__attempt_login()

    @Gtk.Template.Callback()
    def _open_faq(self, _button: Gtk.Button) -> None:
        app = Gio.Application.get_default()
        if not app:
            return None
        window = app.get_active_window()
        Gtk.show_uri(
            window,
            "https://gitlab.gnome.org/World/iotas/-/wikis/nextcloud-notes",
            Gdk.CURRENT_TIME,
        )

    @Gtk.Template.Callback()
    def _on_done_clicked(self, _button: Gtk.Button) -> None:
        self.close()

    def __on_close(self) -> None:
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None

        self.__sync_manager.disconnect_by_func(self.__on_initial_sync_finished)

    def __on_login_success(self, storage_success: bool) -> None:
        if storage_success:
            if self._stack.get_visible_child() == self._loading:
                self._loading.set_description("")
                if self.__mode == LoginDialogMode.REAUTHENTICATE:
                    # Translators: Title
                    self._loading.set_title(_("Updating Notes"))
                else:
                    self.__config_manager.first_start = False
                    # Translators: Title
                    self._loading.set_title(_("Performing Initial Transfer"))
        else:
            self._stack.set_visible_child(self._secret_service_failure)

    def __on_initial_sync_finished(self, _obj: SyncManager, _changes: int) -> None:
        self._stack.set_visible_child(self._finished)
        self._toolbar_view.set_reveal_top_bars(False)

        def removeprefix(instr, prefix):
            return instr[len(prefix) :] if instr.startswith(prefix) else instr

        clean_url = removeprefix(self.__server_url, "https://")
        clean_url = removeprefix(clean_url, "http://")

        self._instance_label.set_text(clean_url)

    def __url_entry_valid(self) -> bool:
        valid = False
        entry_text = self.__get_sanitised_url_entry()
        try:
            valid = GLib.uri_is_valid(entry_text, GLib.UriFlags.NONE)
        except GLib.GError:
            pass
        # GLib's Uri.is_valid appears to allow eg. http:someurl.com
        if valid:
            scheme = GLib.uri_peek_scheme(entry_text)
            if scheme not in ("http", "https") or not entry_text.startswith(f"{scheme}://"):
                valid = False
        return valid

    def __wait_and_check_success(self) -> None:
        self.__timeout_id = GLib.timeout_add(self.CHECK_LOGIN_INTERVAL, self.__check_success)

    def __check_success(self) -> None:
        self.__timeout_id = None

        def thread_do() -> None:
            assert self.__endpoint
            assert self.__token
            (
                login_success,
                storage_success,
            ) = self.__sync_manager.check_for_login_success(self.__endpoint, self.__token)
            if login_success:
                GLib.idle_add(self.__on_login_success, storage_success)
            else:
                GLib.idle_add(self.__wait_and_check_success)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __attempt_login(self) -> None:
        self._stack.set_visible_child(self._loading)
        self._loading.set_description("")
        # Translators: Title
        self._loading.set_title(_("Connecting"))

        self.__server_url = self.__get_sanitised_url_entry()
        self.__config_manager.nextcloud_endpoint = self.__server_url

        def thread_do() -> None:
            # Using instead of getting parent due to __attempt_login being called in constructor
            # for reauthentication
            app = Gio.Application.get_default()
            window = app.get_active_window()
            result, endpoint, token = self.__sync_manager.start_login(self.__server_url, window)
            GLib.idle_add(self.__handle_login_initialisation, result, endpoint, token)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()

    def __handle_login_initialisation(
        self, result: LoginRequestResult, endpoint: Optional[str], token: Optional[str]
    ) -> None:
        if result == LoginRequestResult.SUCCESS:
            # Translators: Title
            self._loading.set_title(_("Waiting for Login"))
            # Translators: Description
            self._loading.set_description(_("Complete the authentication in your browser"))
            self.__endpoint = endpoint
            self.__token = token
            self.__wait_and_check_success()
        elif result == LoginRequestResult.FAILURE_SELF_SIGNED_SSL:
            self._stack.set_visible_child(self._self_signed_certificate_handling)
        else:
            self._stack.set_visible_child(self._entry)
            # Set the failed URL back into try text entry. This is to make any URI scheme
            # prefixing that has been done clear to the user.
            self._url_entry.set_text(self.__server_url)
            self._url_entry.set_position(-1)
            self._url_entry.grab_focus()

            if result == LoginRequestResult.FAILURE_CERTIFICATE:
                # Translators: Description, notification
                msg = _("Failed to start login with possible certificate issue")
            else:
                # Translators: Description, notification
                msg = _("Failed to start login. Wrong address?")

            toast = Adw.Toast.new(msg)
            self._entry.add_toast(toast)

    def __get_sanitised_url_entry(self) -> str:
        entry_text = self._url_entry.get_text().strip().strip("/")
        scheme = GLib.uri_peek_scheme(entry_text)
        if entry_text != "" and scheme is None:
            entry_text = f"https://{entry_text}"
        return entry_text



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject

from abc import abstractmethod
from typing import Optional

from iotas.note import Note


class NoteExporter(GObject.Object):
    """Single note exporter."""

    # __gsignals__ = {
    #     # str: single export out path
    #     "finished": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    #     # str: reason
    #     "failed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    # }

    @abstractmethod
    def export(
        self,
        note: Note,
        out_format: str,
        file_extension: str,
        supporting_tex: bool,
        allow_missing_images: bool = False,
        user_location: Optional[str] = None,
    ) -> None:
        """Export note.

        :param Note note: Note to render
        :param str out_format: Export format
        :param str file_extension: File extension
        :param bool supporting_tex: TeX support
        :param bool allow_missing_images: Whether missing images are treated as a failure
        :param str user_location: User chosen export location, optional
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import re
from typing import Optional


def check_line_for_ordered_list_item(line: str) -> Optional[re.Match]:
    """Check whether string representing line of text contains an ordered list item.

    :param str line: Line of text to search
    :return: Regular expression match object, if matched
    :rtype: re.Match, optional
    """
    term = r"^(\s*)([a-zA-Z]{1}|[0-9]+)([\.\)]){1}[ ]+"
    return re.search(term, line)


def calculate_ordered_list_index(reference: str, direction: int) -> Optional[str]:
    """Calculate an item in an ordered list sequence.

    :param str reference: The element to start from
    :param int direction: The direction, 1 or -1
    :return: The item
    :rtype: str, optional
    """
    sequence_next = None
    if reference.isdigit():
        sequence_next = str(int(reference) + direction)
        if sequence_next == 0:
            sequence_next = None
    else:
        try:
            char_ord = ord(reference)
        except TypeError:
            pass
        else:
            if direction > 0 and reference.upper() != "Z":
                sequence_next = chr(char_ord + 1)
            elif direction < 0 and reference.upper() != "A":
                sequence_next = chr(char_ord - 1)
    return sequence_next


def format_ordered_list_item(spacing: str, index: str, delimiter: str) -> str:
    """Generate string for list item.

    :param str spacing: Pre marker spacing
    :param str index: Item index
    :param str delimiter: The delimiter before the item text
    :return: The built string
    :rtype: str
    """
    return f"{spacing}{index}{delimiter} "



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _

from gi.repository import Adw, Gio, GObject, Gtk

from iotas.category_header_bar import CategoryHeaderBar
from iotas.category_treeview_list_store import CategoryTreeViewListStore
from iotas.note import Note
from iotas.ui_utils import show_error_dialog


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/selection_header_bar.ui")
class SelectionHeaderBar(Adw.Bin):
    __gtype_name__ = "SelectionHeaderBar"

    __gsignals__ = {
        "set-favourite": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "clear-favourite": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "export": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "delete": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "categories-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "abort": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "select-all": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    _stack: Gtk.Stack = Gtk.Template.Child()
    _main_header_bar: Adw.HeaderBar = Gtk.Template.Child()
    _category_header_bar: CategoryHeaderBar = Gtk.Template.Child()
    _count_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self.__active = False
        self.__selected: list[Note] = []
        self._category_header_bar.connect(
            "categories-changed", lambda _o: self.emit("categories-changed")
        )
        self._category_header_bar.connect("abort", lambda _o: self.emit("abort"))
        self.__setup_actions()

    def setup(self, model: CategoryTreeViewListStore) -> None:
        """Perform initial setup.

        :param CategoryTreeViewListStore model: Categories model
        """
        self._category_header_bar.setup(model)

    def activate(self) -> None:
        """Activate header bar."""
        self._stack.set_visible_child(self._main_header_bar)
        self.active = True
        self.__selected = []
        self.__refresh()

    def set_note_selected(self, note: Note, select: bool) -> None:
        """Add or remove note from selection.

        :param Note note: Note to select
        :param bool select: Whether to add or remote from selection
        """
        if select:
            if note not in self.__selected:
                self.__selected.append(note)
        else:
            if note in self.__selected:
                self.__selected.remove(note)
        self.__refresh()

    def get_selected(self) -> list[Note]:
        """Get selected notes.

        :return: Selected notes
        :rtype: list[Note]
        """
        return list(self.__selected)

    def get_categories_changeset(self) -> list[tuple[Note, str]]:
        """Get categories changeset.

        :return: Selected of category-changed notes and their previous categories
        :rtype: list[tuple[Note, str]]
        """
        return self._category_header_bar.take_changeset()

    def edit_category_for_selection(self) -> None:
        """Edit the category for the selected notes."""
        if not self.__selected:
            return

        msg = _("Unable to change category on read-only note")
        if self.__error_for_readonly_in_selection(msg):
            return

        self._stack.set_visible_child(self._category_header_bar)
        self._category_header_bar.activate(self.__selected)

    def toggle_favourite_on_selection(self) -> None:
        """Toggle favourites for the selected notes.

        If a collection of notes with mixed favourite states are selected the destination state
        is the less common one in those selected.
        """
        if not self.__selected:
            return

        set = len([x for x in self.__selected if x.favourite])
        unset = len(self.__selected) - set
        if set >= unset:
            self.emit("clear-favourite")
        else:
            self.emit("set-favourite")

    def export_selection(self) -> None:
        """Export selection."""
        if not self.__selected:
            return

        self.emit("export")

    def handle_delete_keyboard_shortcut(self) -> None:
        """Handle pressed delete keyboard shortcut."""
        if self._stack.get_visible_child() == self._category_header_bar:
            self._category_header_bar.clear_and_apply()
        else:
            self.__delete_selection()

    def select_all_category_text(self) -> None:
        """Select all text in category header bar text entry."""
        if not self.active:
            return
        if self._stack.get_visible_child() == self._category_header_bar:
            self._category_header_bar.select_all_text()

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value

    @GObject.Property(type=bool, default=False)
    def category_entry_focused(self) -> bool:
        if not self.active:
            return False
        if self._stack.get_visible_child() == self._category_header_bar:
            return self._category_header_bar.entry_focused
        else:
            return False

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()

        action = Gio.SimpleAction.new("categorise")
        action.connect("activate", lambda _a, _p: self.edit_category_for_selection())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("favourite")
        action.connect("activate", lambda _a, _p: self.toggle_favourite_on_selection())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("export")
        action.connect("activate", lambda _a, _p: self.export_selection())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("delete")
        action.connect("activate", lambda _a, _p: self.__delete_selection())
        action_group.add_action(action)

        self.insert_action_group("selection", action_group)
        self.__action_group = action_group

    def __delete_selection(self) -> None:
        """Delete the selected notes."""
        if not self.__selected:
            return
        self.emit("delete")

    def __refresh(self) -> None:
        # Translators: Description, {} is a number
        self._count_label.set_text(_("{} Selected").format(len(self.__selected)))

        if self._stack.get_visible_child() == self._category_header_bar:
            self._category_header_bar.replace_notes(self.__selected)

        for action in self.__action_group.list_actions():
            action = self.__action_group.lookup(action)
            action.set_property("enabled", self.__selected)

    def __error_for_readonly_in_selection(self, message: str) -> bool:
        have_readonly = False
        for note in self.__selected:
            if note.read_only:
                have_readonly = True
                break
        if have_readonly:
            # Translators: Description, alert
            show_error_dialog(self, _("Unable to change category on read-only note"))
        return have_readonly



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gdk, GObject, Gtk

from enum import IntEnum, auto

from iotas.index_note_list import IndexNoteList
from iotas.note_manager import NoteManager
from iotas.ui_utils import check_for_search_starting, check_for_open_first_result_shortcut


class IndexSearchState(IntEnum):
    NO_TERM = auto()
    EMPTY = auto()
    RESULTS = auto()


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/index_search_header_bar.ui")
class IndexSearchHeaderBar(Adw.Bin):
    __gtype_name__ = "IndexSearchHeaderBar"
    __gsignals__ = {
        "changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    _entry: Gtk.SearchEntry = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self.__active = False

    def setup(self, note_list: IndexNoteList, note_manager: NoteManager) -> None:
        """Perform initial setup.

        :param IndexNoteList note_list: The note list
        :param NoteManager note_manager: The note manager
        """
        self.__note_list = note_list
        self.__note_manager = note_manager

    def check_if_starting(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        state: Gdk.ModifierType,
        clear: bool = True,
    ) -> bool:
        """Check if starting search via type to search.

        :param Gtk.EventControllerKey controller: The key controller
        :param int keyval: The key value
        :param Gdk.ModifierType state: Any modifier state
        :param bool clear: Whether to clear the entry
        :return: Whether search is starting
        :rtype: bool
        """
        if not check_for_search_starting(controller, keyval, state):
            return False

        if clear:
            self._entry.set_text("")
        if controller.forward(self._entry.get_delegate()):
            return True

        return False

    def check_for_open_first_result_shortcut(
        self, controller: Gtk.EventControllerKey, keyval: int, state: Gdk.ModifierType
    ) -> bool:
        """Check keyboard shortcut to activate first result has been pressed.

        :param Gtk.EventControllerKey controller: The key controller
        :param int keyval: The key value
        :param Gdk.ModifierType state: Any modifier state
        :return: Whether pressed
        :rtype: bool
        """
        # Clunky way to determine if we have input focus
        window = self.get_root()
        focus_widget = window.get_property("focus-widget")
        if focus_widget.is_ancestor(self._entry):
            return check_for_open_first_result_shortcut(keyval, state)
        else:
            return False

    def enter(self, clear_text: bool, select_all: bool) -> None:
        """Enter search.

        :param bool clear_text: Whether to clear existing text
        :param bool select_all: Whether to select all text (when not clearing)
        """
        self._entry.grab_focus()
        if clear_text:
            self._entry.set_text("")
        elif select_all:
            self._entry.select_region(0, -1)
        self.active = True

    def search(self, show_all: bool) -> IndexSearchState:
        """Perform a search.

        :param bool show_all: Whether to show the full result set (instead of the most recent with
            an expandable section)
        :return: The resulting search state
        :rtype: IndexSearchState
        """
        search_text = self.text
        if len(search_text) > 0:
            found_ids = self.__note_manager.search_notes(search_text, sort=True)
            if len(found_ids) > 0:
                self.__note_list.restrict_for_search_by_ids(found_ids, show_all)
                return IndexSearchState.RESULTS
            else:
                return IndexSearchState.EMPTY
        else:
            return IndexSearchState.NO_TERM

    def exit(self) -> None:
        """Exit search.

        :param bool value: New value
        """
        self.active = False

    def focus_entry(self) -> None:
        """Focus the text entry."""
        self._entry.grab_focus()

    def select_all_text(self) -> None:
        """Select all entry text"""
        self._entry.select_region(0, -1)

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value

    @GObject.Property(type=str)
    def text(self) -> str:
        return self._entry.get_text().strip()

    @text.setter
    def set_text(self, value: str) -> str:
        return self._entry.set_text(value)

    @GObject.Property(type=bool, default=False)
    def entry_focused(self) -> bool:
        return self._entry.has_focus

    @Gtk.Template.Callback()
    def _on_entry_changed(self, _entry: Gtk.Editable) -> None:
        self.emit("changed")



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from typing import Callable
import unittest
import warnings

from gi.repository import GLib, Gtk, GtkSource
from iotas.formatter import Formatter, LinkStateInfo

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    formatter: Formatter
    buffer: GtkSource.Buffer
    awaiting_signal = False

    def test_bold(self) -> None:
        # Markup around cursor
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.bold)
        self.assertEqual(self.__get_text(), "****")
        centre = buffer.get_iter_at_offset(2)
        buffer.place_cursor(centre)
        formatter.bold()
        self.assertEqual(self.__get_text(), "")

        # Markup around selection
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "text")
        buffer.select_range(buffer.get_start_iter(), start_iter)
        self.__call_and_wait(formatter.bold)
        self.assertEqual(self.__get_text(), "**text**")
        self.__call_and_wait(formatter.bold)
        self.assertEqual(self.__get_text(), "text")
        Gtk.TextBuffer.select_range

        # Removing markup when cursor within bold
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.bold)
        self.assertEqual(self.__get_text(), "**text**")
        buffer.place_cursor(buffer.get_iter_at_offset(4))
        self.__call_and_wait(formatter.bold)
        self.assertEqual(self.__get_text(), "text")

        # Multi-line selection ignored (uses selection start as cursor instead)
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "line\nline\nline\nline")
        buffer.select_range(buffer.get_start_iter(), buffer.get_iter_at_offset(15))
        self.__call_and_wait(formatter.bold)
        self.assertEqual(self.__get_text(), "****line\nline\nline\nline")

    def test_italic(self) -> None:
        # Markup around cursor
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.italic)
        self.assertEqual(self.__get_text(), "__")
        centre = buffer.get_iter_at_offset(1)
        buffer.place_cursor(centre)
        formatter.italic()
        self.assertEqual(self.__get_text(), "")

        # Markup around selection
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "text")
        buffer.select_range(buffer.get_start_iter(), start_iter)
        self.__call_and_wait(formatter.italic)
        self.assertEqual(self.__get_text(), "_text_")
        self.__call_and_wait(formatter.italic)
        self.assertEqual(self.__get_text(), "text")
        Gtk.TextBuffer.select_range

        # Removing markup when cursor within italic
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.italic)
        self.assertEqual(self.__get_text(), "_text_")
        buffer.place_cursor(buffer.get_iter_at_offset(3))
        self.__call_and_wait(formatter.italic)
        self.assertEqual(self.__get_text(), "text")

        # Multi-line selection ignored (uses selection start as cursor instead)
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "line\nline\nline\nline")
        buffer.select_range(buffer.get_start_iter(), buffer.get_iter_at_offset(15))
        self.__call_and_wait(formatter.italic)
        self.assertEqual(self.__get_text(), "__line\nline\nline\nline")

    def test_strikethrough(self) -> None:
        # Markup around cursor
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.strikethrough)
        self.assertEqual(self.__get_text(), "~~~~")
        centre = buffer.get_iter_at_offset(2)
        buffer.place_cursor(centre)
        formatter.strikethrough()
        self.assertEqual(self.__get_text(), "")

        # Markup around selection
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "text")
        buffer.select_range(buffer.get_start_iter(), start_iter)
        self.__call_and_wait(formatter.strikethrough)
        self.assertEqual(self.__get_text(), "~~text~~")
        self.__call_and_wait(formatter.strikethrough)
        self.assertEqual(self.__get_text(), "text")
        Gtk.TextBuffer.select_range

        # Removing markup when cursor within strikethrough
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.strikethrough)
        self.assertEqual(self.__get_text(), "~~text~~")
        buffer.place_cursor(buffer.get_iter_at_offset(4))
        self.__call_and_wait(formatter.strikethrough)
        self.assertEqual(self.__get_text(), "text")

        # Multi-line selection ignored (uses selection start as cursor instead)
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "line\nline\nline\nline")
        buffer.select_range(buffer.get_start_iter(), buffer.get_iter_at_offset(15))
        self.__call_and_wait(formatter.strikethrough)
        self.assertEqual(self.__get_text(), "~~~~line\nline\nline\nline")

    def test_quote(self) -> None:
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "> ")
        formatter.quote()
        self.assertEqual(self.__get_text(), "")

        # Cursor location and selection not making a difference
        buffer.insert(buffer.get_start_iter(), "text")
        buffer.place_cursor(self.buffer.get_end_iter())
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "> text")
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "text")
        buffer.place_cursor(self.buffer.get_start_iter())
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "> text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "text")
        buffer.select_range(buffer.get_iter_at_offset(2), buffer.get_iter_at_offset(3))
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "> text")

        # Basic multi-line
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "line\nline\nline\nline")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "> line\n> line\n> line\n> line")
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "line\nline\nline\nline")

        # Mixed quote indentation multi-line
        formatter, buffer = self.__reset()
        start_iter = buffer.get_start_iter()
        buffer.insert(start_iter, "line\n> line\nline\nline")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "> line\n> > line\n> line\n> line")
        self.__call_and_wait(formatter.quote)
        self.assertEqual(self.__get_text(), "line\n> line\nline\nline")

    def test_heading(self) -> None:
        # Empty
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.heading, 1)
        self.assertEqual(self.__get_text(), "# ")
        formatter.heading(0)
        self.assertEqual(self.__get_text(), "")

        # Basic text
        buffer.insert(buffer.get_start_iter(), "text")
        self.__call_and_wait(formatter.heading, 1)
        self.assertEqual(self.__get_text(), "# text")
        self.__call_and_wait(formatter.heading, 3)
        self.assertEqual(self.__get_text(), "### text")
        formatter.heading(0)
        self.assertEqual(self.__get_text(), "text")

        # Cursor location and selection not making a difference
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "text")
        buffer.place_cursor(self.buffer.get_end_iter())
        self.__call_and_wait(formatter.heading, 1)
        self.assertEqual(self.__get_text(), "# text")
        formatter.heading(0)
        self.assertEqual(self.__get_text(), "text")
        buffer.place_cursor(self.buffer.get_start_iter())
        self.__call_and_wait(formatter.heading, 1)
        self.assertEqual(self.__get_text(), "# text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        formatter.heading(0)
        self.assertEqual(self.__get_text(), "text")
        buffer.select_range(buffer.get_iter_at_offset(2), buffer.get_iter_at_offset(3))
        self.__call_and_wait(formatter.heading, 1)
        self.assertEqual(self.__get_text(), "# text")

    def test_get_link_state(self) -> None:
        # Empty
        formatter, buffer = self.__reset()
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "")
        self.assertEqual(state.text, "")

        # Typical inline link
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "[text](<https://linky.com>)")
        buffer.place_cursor(buffer.get_start_iter())
        state = formatter.get_link_state()
        self.assertFalse(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "https://linky.com")
        self.assertEqual(state.text, "text")
        self.assertEqual(state.source_text, "[text](<https://linky.com>)")
        self.assertEqual(state.start_offset, 0)
        self.assertEqual(state.end_offset, 27)

        # Creating inline from automatic
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "<https://linky.com>")
        buffer.place_cursor(buffer.get_start_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertTrue(state.automatic_link)
        self.assertEqual(state.link, "https://linky.com")
        self.assertEqual(state.text, "")

        # URL in plain text
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "https://linky.com")
        buffer.place_cursor(buffer.get_start_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "https://linky.com")
        self.assertEqual(state.text, "")

        # New from URL selection
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "https://linky.com")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "https://linky.com")
        self.assertEqual(state.text, "")

        # Edit selected inline
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "[text](<https://linky.com>)")
        buffer.select_range(buffer.get_iter_at_offset(3), buffer.get_iter_at_offset(10))
        state = formatter.get_link_state()
        self.assertFalse(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "https://linky.com")
        self.assertEqual(state.text, "text")
        self.assertEqual(state.source_text, "[text](<https://linky.com>)")
        self.assertEqual(state.start_offset, 0)
        self.assertEqual(state.end_offset, 27)

        # Creating inline from selected automatic
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "<https://linky.com>")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertTrue(state.automatic_link)
        self.assertEqual(state.link, "https://linky.com")
        self.assertEqual(state.text, "")

        # Selection text for title in creation
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "plain text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "")
        self.assertEqual(state.text, "plain text")

        # Avoiding creating link from selection which has markup
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- item")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "")
        self.assertEqual(state.text, "")

        # Create when seeing only part of URI
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "plain text https://")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "")
        self.assertEqual(state.text, "")

        # Ignore selection when multi-line, create new
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "[text](<https://linky.com>)\ntext")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        state = formatter.get_link_state()
        self.assertTrue(state.creating)
        self.assertFalse(state.automatic_link)
        self.assertEqual(state.link, "")
        self.assertEqual(state.text, "")

    def test_link(self) -> None:
        # No replacement
        formatter, buffer = self.__reset()
        info = LinkStateInfo()
        info.text = "text"
        info.link = "https://linky.com"
        formatter.link(info)
        self.assertEqual(self.__get_text(), "[text](https://linky.com)")

        # Replace text used for title
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "plain text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        info = formatter.get_link_state()
        self.assertTrue(info.creating)
        self.assertFalse(info.automatic_link)
        self.assertEqual(info.link, "")
        self.assertEqual(info.text, "plain text")
        info.link = "https://linky.com"
        formatter.link(info)
        self.assertEqual(self.__get_text(), "[plain text](https://linky.com)")

        # Edit inline
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "[text](https://linky.com)")
        buffer.place_cursor(buffer.get_start_iter())
        info = formatter.get_link_state()
        self.assertFalse(info.creating)
        self.assertFalse(info.automatic_link)
        self.assertEqual(info.link, "https://linky.com")
        self.assertEqual(info.text, "text")
        self.assertEqual(info.source_text, "[text](https://linky.com)")
        self.assertEqual(info.start_offset, 0)
        self.assertEqual(info.end_offset, 25)
        info.link = "https://newlink.com"
        info.text = "new text"
        formatter.link(info)
        self.assertEqual(self.__get_text(), "[new text](https://newlink.com)")

        # Insert instead of edit if buffer is updated remotely while in edit link dialog
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "[text](https://linky.com)")
        buffer.place_cursor(buffer.get_start_iter())
        info = formatter.get_link_state()
        info.link = "https://newlink.com"
        info.text = "new text"
        self.__insert_and_wait(buffer.get_start_iter(), "modified")
        formatter.link(info)
        self.assertEqual(
            self.__get_text(), "modified[new text](https://newlink.com)[text](https://linky.com)"
        )

    def test_code(self) -> None:
        # Markup around cursor
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.code)
        self.assertEqual(self.__get_text(), "``")
        formatter.code()
        self.assertEqual(self.__get_text(), "")

        # Code block
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "line\nline\nline")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.code)
        self.assertEqual(self.__get_text(), "```\nline\nline\nline\n```\n")
        formatter.code()
        self.assertEqual(self.__get_text(), "line\nline\nline\n\n")

        # Basic span
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "code")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.code)
        self.assertEqual(self.__get_text(), "`code`")
        formatter.code()
        self.assertEqual(self.__get_text(), "code")

        # Double backtick span
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "``code``")
        buffer.place_cursor(buffer.get_iter_at_offset(3))
        formatter.code()
        self.assertEqual(self.__get_text(), "code")

    def test_unordered_list(self) -> None:
        # Empty
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.unordered_list)
        self.assertEqual(self.__get_text(), "- ")
        formatter.unordered_list()
        self.assertEqual(self.__get_text(), "")

        # Basic
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "text")
        self.__call_and_wait(formatter.unordered_list)
        self.assertEqual(self.__get_text(), "- text")

        # Multi-line
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "text\ntext\n- text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.unordered_list)
        self.assertEqual(self.__get_text(), "- text\n- text\ntext")
        self.__call_and_wait(formatter.unordered_list)
        self.assertEqual(self.__get_text(), "text\ntext\n- text")

        # Other list items
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- [ ] text")
        self.__call_and_wait(formatter.unordered_list)
        self.assertEqual(self.__get_text(), "- text")
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "1. text")
        self.__call_and_wait(formatter.unordered_list)
        self.assertEqual(self.__get_text(), "- text")

    def test_ordered_list(self) -> None:
        # Empty
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "1. ")
        formatter.ordered_list()
        self.assertEqual(self.__get_text(), "")

        # Basic
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "text")
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "1. text")

        # Multi-line
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "text\ntext\ntext")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "1. text\n2. text\n3. text")
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "text\ntext\ntext")

        # Continuation of existing
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "1. text\n2. text\n3. text\ntext\ntext")
        success, line_iter = buffer.get_iter_at_line(3)
        self.assertTrue(success)
        buffer.select_range(line_iter, buffer.get_end_iter())
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "1. text\n2. text\n3. text\n4. text\n5. text")

        # Other list items
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- [ ] text\n* text\n+ text\n- text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "1. text\n2. text\n3. text\n4. text")
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- [ ] text")
        self.__call_and_wait(formatter.ordered_list)
        self.assertEqual(self.__get_text(), "1. text")

    def test_checkbox(self) -> None:
        # Empty
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.checkbox)
        self.assertEqual(self.__get_text(), "- [ ] ")
        formatter.checkbox()
        self.assertEqual(self.__get_text(), "")

        # Basic
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "text")
        self.__call_and_wait(formatter.checkbox)
        self.assertEqual(self.__get_text(), "- [ ] text")

        # Multi-line
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "text\ntext\n- [ ] text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.checkbox)
        self.assertEqual(self.__get_text(), "- [ ] text\n- [ ] text\ntext")
        self.__call_and_wait(formatter.checkbox)
        self.assertEqual(self.__get_text(), "text\ntext\n- [ ] text")

        # Other list items
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- text\n* text\n+ text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.checkbox)
        self.assertEqual(self.__get_text(), "- [ ] text\n- [ ] text\n- [ ] text")
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "1. text")
        self.__call_and_wait(formatter.checkbox)
        self.assertEqual(self.__get_text(), "- [ ] text")

    def test_toggle_checkbox(self) -> None:
        # Empty
        formatter, buffer = self.__reset()
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "")

        # Plain text
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "text")
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "text")

        # Basic
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- [ ] text")
        self.__call_and_wait(formatter.toggle_checkbox)
        self.assertEqual(self.__get_text(), "- [x] text")
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "- [ ] text")

        # Indented
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "  - [ ] text")
        self.__call_and_wait(formatter.toggle_checkbox)
        self.assertEqual(self.__get_text(), "  - [x] text")
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "  - [ ] text")

        # Multi line
        formatter, buffer = self.__reset()
        self.__insert_and_wait(buffer.get_start_iter(), "- [ ] text\n- [ ] text\n- [x] text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        self.__call_and_wait(formatter.toggle_checkbox)
        self.assertEqual(self.__get_text(), "- [x] text\n- [x] text\n- [ ] text")
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "- [ ] text\n- [ ] text\n- [x] text")

        # Other list items
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "- text")
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "- text")
        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "1. text")
        formatter.toggle_checkbox()
        self.assertEqual(self.__get_text(), "1. text")

    def test_horizontal_rule(self) -> None:
        formatter, buffer = self.__reset()
        self.__call_and_wait(formatter.horizontal_rule)
        self.assertEqual(self.__get_text(), "\n- - -\n")
        self.__call_and_wait(formatter.horizontal_rule)
        self.assertEqual(self.__get_text(), "\n\n")

        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "text")
        buffer.place_cursor(buffer.get_start_iter())
        self.__call_and_wait(formatter.horizontal_rule)
        self.assertEqual(self.__get_text(), "\n- - -\n\ntext")

        formatter, buffer = self.__reset()
        buffer.insert(buffer.get_end_iter(), "text")
        buffer.place_cursor(buffer.get_start_iter())
        self.__call_and_wait(formatter.horizontal_rule)
        self.assertEqual(self.__get_text(), "\n- - -\n\ntext")

    def test_table(self) -> None:

        def add_table(columns, rows, view_width, character_width) -> None:
            self.__call_and_wait(formatter.table, columns, rows, view_width, character_width)

        def check_table(columns: int, rows: int, min_column_width: int) -> bool:
            text = self.__get_text()
            lines = text.strip().split("\n")
            if len(lines) != rows + 2:
                return False
            if lines[0].replace(" ", "") != "|" * (columns + 1):
                return False
            return True

        formatter, buffer = self.__reset()
        add_table(columns=3, rows=3, view_width=600, character_width=5)
        self.assertTrue(check_table(3, 3, 20))

        # Check GFM min column width
        formatter, buffer = self.__reset()
        add_table(columns=300, rows=3, view_width=600, character_width=5)
        self.assertTrue(check_table(300, 3, 3))

    def __get_language(self) -> tuple[GtkSource.LanguageManager, GtkSource.Language]:
        manager = GtkSource.LanguageManager()
        file_dir = os.path.dirname(__file__)
        language_path = os.path.join(
            file_dir, os.pardir, "data", "gtksourceview-5", "language-specs"
        )
        manager.prepend_search_path(language_path)
        language = manager.get_language("iotas-markdown")
        return (manager, language)

    def __reset(self) -> tuple[Formatter, GtkSource.Buffer]:
        self.formatter = Formatter()
        self.__manager, language = self.__get_language()
        self.buffer = GtkSource.Buffer.new_with_language(language)
        self.formatter.buffer = self.buffer
        return (self.formatter, self.buffer)

    def __get_text(self) -> str:
        if self.buffer is None:
            return ""
        else:
            start = self.buffer.get_start_iter()
            end = self.buffer.get_end_iter()
            return self.buffer.get_text(start, end, False)

    def __call_and_wait(self, call: Callable, *args) -> None:
        context = GLib.MainContext.default()
        self.buffer.connect("highlight-updated", self.__flag_not_waiting)
        self.awaiting_signal = True
        call(*args)
        while self.awaiting_signal:
            context.iteration(False)

    def __insert_and_wait(self, location: Gtk.TextIter, text: str) -> None:
        context = GLib.MainContext.default()
        self.buffer.connect("highlight-updated", self.__flag_not_waiting)
        self.awaiting_signal = True
        self.buffer.insert(location, text)
        while self.awaiting_signal:
            context.iteration(False)

    def __flag_not_waiting(self, *_args) -> None:
        self.awaiting_signal = False



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/index_menu_button.ui")
class IndexMenuButton(Gtk.MenuButton):
    __gtype_name__ = "IndexMenuButton"



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import warnings

from markdown_it import MarkdownIt

from iotas.markdown_helpers import (
    parse_to_tokens,
    get_note_export_content,
    get_image_attachments_from_tokens,
    get_image_attachments_from_note_content,
    filter_image_tokens,
    create_attachment_for_markup_path,
    update_image_path_in_markup,
    string_is_image,
)
from iotas.note import Note

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_parse_to_tokens(self) -> None:
        note = Note()
        note.content = "plain"
        parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        self.assertIsInstance(parser, MarkdownIt)
        self.assertEqual(len(tokens), 3)

    def test_get_note_export_content(self) -> None:
        note = Note()
        note.content = "![description](image.png)"
        note.id = 1
        out = get_note_export_content(note, prefix_note_id=False)
        self.assertEqual(out, "![description](attachments/image.png)")

        out = get_note_export_content(note, prefix_note_id=True)
        self.assertEqual(out, "![description](attachments/1.image.png)")

    def test_get_image_attachments_from_tokens(self) -> None:
        note = Note()

        note.content = "plain"
        _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        attachments = get_image_attachments_from_tokens(note, tokens)
        self.assertEqual(len(attachments), 0)

        note.content = "![](image.png)\n\n![](image2.png)"
        _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        attachments = get_image_attachments_from_tokens(note, tokens)
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].path, "image.png")
        self.assertEqual(attachments[1].path, "image2.png")

    def test_get_image_attachments_from_note_content(self) -> None:
        note = Note()

        note.content = "plain"
        attachments = get_image_attachments_from_note_content(note)
        self.assertEqual(len(attachments), 0)

        note.content = "![](image.png)\n\n![](image2.png)"
        attachments = get_image_attachments_from_note_content(note)
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].path, "image.png")
        self.assertEqual(attachments[1].path, "image2.png")

        note.content = '![description](image.png "title")'
        attachments = get_image_attachments_from_note_content(note)
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0].path, "image.png")

    def test_filter_image_tokens(self) -> None:
        note = Note()

        note.content = "plain"
        _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        tokens = filter_image_tokens(tokens)
        self.assertEqual(len(tokens), 0)

        note.content = "![](image.png)\n\n![](image2.png)"
        _parser, tokens = parse_to_tokens(note, exporting=False, tex_support=False)
        tokens = filter_image_tokens(tokens)
        self.assertEqual(len(tokens), 2)

    def test_create_attachment_for_markup_path(self) -> None:
        note = Note()
        note.id = 1
        note.remote_id = 2
        attachment = create_attachment_for_markup_path(note, "image.png")
        self.assertEqual(attachment.note_id, 1)
        self.assertEqual(attachment.note_remote_id, 2)
        self.assertEqual(attachment.path, "image.png")

    def test_update_image_path_in_markup(self) -> None:
        content = "# basic"
        out = update_image_path_in_markup(content, "old_path.png", "new_path.png")
        self.assertEqual(out, content)

        content = "![description](old.png)"
        out = update_image_path_in_markup(content, "old.png", "new.png")
        self.assertEqual(out, "![description](new.png)")

        content = "![](old.png"
        out = update_image_path_in_markup(content, "old.png", "new.png")
        self.assertEqual(out, "![](old.png")

        content = '![](old.png "title")'
        out = update_image_path_in_markup(content, "old.png", "new.png")
        self.assertEqual(out, '![](new.png "title")')

        content = "[](old.png)"
        out = update_image_path_in_markup(content, "old.png", "new.png")
        self.assertEqual(out, "[](old.png)")

    def test_string_is_image(self) -> None:
        self.assertFalse(string_is_image("plain"))
        self.assertTrue(string_is_image("![](image.png)"))
        self.assertFalse(string_is_image("![](image.png"))
        self.assertFalse(string_is_image("[](image.png)"))
        self.assertFalse(string_is_image(" ![](image.png)"))
        self.assertFalse(string_is_image("![](image.png) "))
        self.assertTrue(string_is_image("![description](image.png)"))
        self.assertTrue(string_is_image('![description](image.png "title")'))



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import warnings

from iotas.uri_to_note_creator import UriToNoteCreator, UriParseResult
from iotas.note import Note

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_parse(self) -> None:
        creator = UriToNoteCreator()

        self.assertFalse(creator.parse("plain").success)
        self.assertFalse(creator.parse("uri://text").success)
        self.assertFalse(creator.parse("iotas-notes://text").success)
        self.assertFalse(creator.parse("iotas-notes://?content=text").success)
        self.assertFalse(creator.parse("iotas-notes://create-note?content=text").success)
        self.assertFalse(creator.parse("iotas-notes://create-note?title=text").success)

        result = creator.parse("iotas-notes://create-note?content=text&title=title")
        self.assertIsInstance(result, UriParseResult)
        self.assertTrue(result.success)
        self.assertIsNotNone(result.note)
        assert result.note is not None  # mypy
        self.assertIsInstance(result.note, Note)
        self.assertEqual(result.note.content, "text")
        self.assertEqual(result.note.title, "title")
        self.assertFalse(result.open)

        result = creator.parse("iotas-notes://create-note?content=text&title=title&open=1")
        self.assertTrue(result.success)
        self.assertIsNotNone(result.note)
        assert result.note is not None  # mypy
        self.assertEqual(result.note.content, "text")
        self.assertEqual(result.note.title, "title")
        self.assertTrue(result.open)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib

import os
import pytest
import shutil
from typing import Callable, Optional
import unittest
import warnings

from markdown_it.token import Token
from markdown_it.utils import EnvType, OptionsDict

import pypandoc

from iotas.attachment import Attachment
import iotas.attachment_helpers
from iotas.default_note_exporter import DefaultNoteExporter
from iotas.html_generator import HtmlGenerator
from iotas.note import Note
from iotas.pdf_exporter import PdfExporter


def get_output_dir() -> str:
    file_dir = os.path.dirname(__file__)
    return os.path.join(file_dir, os.pardir, "testing-tmp")


def get_attachments_dir() -> str:
    return os.path.join(get_output_dir(), "user-attachments")


iotas.attachment_helpers.get_attachments_dir = get_attachments_dir

warnings.filterwarnings("ignore", "version")


class _MockPdfExporter(PdfExporter):

    note: Note
    location: str

    def set_callbacks(self, finished_callback: Callable, error_callback: Callable) -> None:
        """Set functions to be called upon export result.

        :param Callable finished_callback: Finished callback
        :param Callable error_callback: Error callback
        """
        pass

    def export(self, note: Note, location: str) -> None:
        """Export PDF of note.

        :param Note note: Note to export
        :param str location: Destination location
        """
        self.note = note
        self.location = location


class MockHtmlGenerator(HtmlGenerator):
    """HTML generator mocked."""

    def generate(
        self,
        note: Note,
        tokens: list[Token],
        render_func: Callable[[list[Token], OptionsDict, EnvType], str],
        parser_options: OptionsDict,
        searching: bool,
        export_format: Optional[str],
        scroll_position: Optional[float] = None,
    ) -> str:
        """Generator HTML for note.

        :param Note note: Note to render
        :param list[Token] tokens: Parser tokens
        :param Callable[[list[Token], OptionsDict, EnvType], str] render_fun: Render function
        :param OptionsDict parser_options: Parser options
        :param bool searching: Whether search CSS should be included
        :param Optional[str] export_format: Export format, if using
        :param Optional[float] scroll_position: Position to scroll to
        :return: Generated HTML
        :rtype: str
        """
        return self.get_output(note)

    def generate_user_stylesheet(self, searching: bool, height: int) -> str:
        """Generate part of stylesheet based on state (configuration etc).

        :param bool searching: Whether searching
        :param int height: View height
        :return: stylesheet
        :rtype: str
        """
        return ""

    def update_font_family(self, family: str) -> None:
        """Update the font family.

        :param str family: New font family
        """
        pass

    def get_output(self, note: Note) -> str:
        """Get the mocked output.

        :param Note note: The note
        :return: Generated content:
        :rtype: str
        """
        return f"{note.title}|{note.content}"


class Test(unittest.TestCase):

    exporter: DefaultNoteExporter
    pdf_exporter: _MockPdfExporter

    def test_export_md_attachments(self) -> None:
        exporter = self.__reset()
        note = Note(new_note=True)
        note.id = 1
        note.content = "![](image.png)"
        note.title = "Test Note Title"

        out_path = self.__prepare_output_dir()

        self.__create_attachment(note, "image.png")

        dir_path = os.path.join(out_path, "Test Note Title")
        file_path = os.path.join(dir_path, "Test Note Title.md")
        self.__call_and_wait(
            exporter.export,
            note,
            out_format="md",
            file_extension="md",
            supporting_tex=False,
            allow_missing_images=False,
            user_location=dir_path,
        )
        self.assertTrue(os.path.exists(file_path))
        with open(file_path, "r") as f:
            contents = f.read()
        self.assertEqual(contents, "![](attachments/image.png)")

        attachment_path = os.path.join(dir_path, "attachments", "image.png")
        self.assertTrue(os.path.exists(attachment_path))

        self.__clean_output_dir()

    def test_export_md_no_attachments(self) -> None:
        exporter = self.__reset()
        note = Note(new_note=True)
        note.id = 1
        note.content = "content"
        note.title = "Test Note Title"

        out_path = self.__prepare_output_dir()

        file_path = os.path.join(out_path, "Test Note Title.md")
        self.__call_and_wait(
            exporter.export,
            note,
            out_format="md",
            file_extension="md",
            supporting_tex=False,
            allow_missing_images=False,
            user_location=file_path,
        )
        self.assertTrue(os.path.exists(file_path))
        with open(file_path, "r") as f:
            contents = f.read()
        self.assertEqual(contents, note.content)

        self.__clean_output_dir()

    def test_export_mock_html(self) -> None:
        exporter = self.__reset()
        note = Note(new_note=True)
        note.id = 1
        note.content = "content\n\n![](image.png)"
        note.title = "Test Note Title"

        out_path = self.__prepare_output_dir()

        self.__create_attachment(note, "image.png")

        dir_path = os.path.join(out_path, "Test Note Title")
        self.__call_and_wait(
            exporter.export,
            note,
            out_format="html",
            file_extension="html",
            supporting_tex=False,
            allow_missing_images=False,
            user_location=dir_path,
        )
        index_path = os.path.join(dir_path, "index.html")
        self.assertTrue(os.path.exists(index_path))
        with open(index_path, "r") as f:
            contents = f.read()
        self.assertTrue(f"{note.content}" in contents)
        self.assertTrue(f"{note.title}" in contents)

        attachment_path = os.path.join(dir_path, "attachments", "image.png")
        self.assertTrue(os.path.exists(attachment_path))

        self.__clean_output_dir()

    def test_mock_export_pdf(self) -> None:
        exporter = self.__reset()
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        out_path = self.__prepare_output_dir()

        file_path = os.path.join(out_path, "Test Note Title.pdf")
        exporter.export(
            note,
            out_format="pdf",
            file_extension="pdf",
            supporting_tex=False,
            allow_missing_images=False,
            user_location=file_path,
        )
        self.assertEqual(self.pdf_exporter.note, note)
        self.assertEqual(self.pdf_exporter.location, file_path)

        self.__clean_output_dir()

    def test_export_odt(self) -> None:
        # TODO Skipped during CI for https://gitlab.gnome.org/World/iotas/-/issues/260
        if "DISPLAY" not in os.environ:
            return pytest.skip("DISPLAY not set")

        exporter = self.__reset()
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        out_path = self.__prepare_output_dir()

        file_path = os.path.join(out_path, "Test Note Title.odt")
        self.__call_and_wait(
            exporter.export,
            note,
            out_format="odt",
            file_extension="odt",
            supporting_tex=False,
            user_location=file_path,
        )
        self.assertTrue(os.path.exists(file_path))
        with open(file_path, "rb") as f:
            contents = f.read()

        md = pypandoc.convert_text(contents, to="md", format="odt")
        self.assertTrue(note.content == md.strip())

        self.__clean_output_dir()

    def __reset(self) -> DefaultNoteExporter:
        self.pdf_exporter = _MockPdfExporter()
        html_generator = MockHtmlGenerator()
        # data_path = "/mock/data/path"
        data_path = "data"
        self.exporter = DefaultNoteExporter(self.pdf_exporter, html_generator, data_path)
        self.exporter.connect("finished", self.__on_finished)
        self.exporter.connect("failed", self.__on_failed)
        return self.exporter

    def __prepare_output_dir(self) -> str:
        out_path = get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)
        os.makedirs(out_path)
        return out_path

    def __clean_output_dir(self) -> None:
        out_path = get_output_dir()
        if os.path.exists(out_path):
            shutil.rmtree(out_path)

    def __create_attachment(self, note: Note, path: str) -> bool:
        os.makedirs(get_attachments_dir())
        attachment = Attachment()
        attachment.note_id = note.id
        attachment.path = path
        return iotas.attachment_helpers.write_attachment_to_disk(attachment, b"bytes")

    def __call_and_wait(self, call: Callable, *args, **kwargs) -> None:
        context = GLib.MainContext.default()
        self.success = False
        self.awaiting_signal = True
        call(*args, **kwargs)
        while self.awaiting_signal:
            context.iteration(False)

    def __on_finished(self, *args) -> None:
        self.awaiting_signal = False
        self.success = True

    def __on_failed(self, _obj: DefaultNoteExporter, error) -> None:
        print(f"Failed: {error}")
        self.awaiting_signal = False



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gdk, GLib, GObject, Gtk

from typing import Callable

from iotas.category import Category, CategorySpecialPurpose
from iotas.category_manager import CategoryManager
from iotas.config_manager import ConfigManager


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/sidebar_row.ui")
class SidebarRow(Gtk.TreeExpander):
    __gtype_name__ = "SidebarRow"

    FOCUS_CSS_CLASS_NAME = "focused"

    _icon: Gtk.Image = Gtk.Template.Child()
    _name_label: Gtk.Label = Gtk.Template.Child()
    _count_with_children_label: Gtk.Label = Gtk.Template.Child()
    _count_node_only_label: Gtk.Label = Gtk.Template.Child()

    def __init__(
        self, activated_func: Callable[[Category, bool], None], category_manager: CategoryManager
    ) -> None:
        super().__init__()
        self.__category: Category | None = None
        self.__bindings: list[GObject.Binding] = []
        self.__activated_func = activated_func
        category_manager.bind_property(
            "have-sub-category", self, "indent-for-icon", GObject.BindingFlags.SYNC_CREATE
        )

        # Add some (hopefully temporary) handling of the keyboard focus ring
        def manage_manual_css_class(row, _flags: Gtk.StateFlags):
            parent = row.get_property("parent")
            if parent is None:
                return
            if row.get_state_flags() & Gtk.StateFlags.FOCUS_VISIBLE:
                parent.add_css_class(self.FOCUS_CSS_CLASS_NAME)
            else:
                parent.remove_css_class(self.FOCUS_CSS_CLASS_NAME)

        self.connect("state-flags-changed", manage_manual_css_class)

    def bind(self, category: Category) -> None:
        """Bind category to row.

        :param Category category: New category
        """
        self.__category = category
        self.__add_binding(category, "note-count", self._count_node_only_label, "label")
        if self.__category.get_parent() is None:
            # Build for root category
            self.__add_binding(category, "display-name", self._name_label, "label")
            self.__add_binding(
                category, "note-count-with-children", self._count_with_children_label, "label"
            )
            tree_list_row = self.get_list_row()
            self.__add_binding(
                tree_list_row,
                "expanded",
                self._count_with_children_label,
                "visible",
                GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.INVERT_BOOLEAN,
            )
            self.__add_binding(tree_list_row, "expanded", self._count_node_only_label, "visible")

            if category.special_purpose == CategorySpecialPurpose.ALL:
                self._icon.set_from_icon_name("edit-select-all-symbolic")
            elif category.special_purpose == CategorySpecialPurpose.UNCATEGORISED:
                self._icon.set_from_icon_name("grid-symbolic")
        else:
            # Build for sub category
            self.__add_binding(category, "sub-category-name", self._name_label, "label")
            self._count_node_only_label.set_visible(True)
            self._count_with_children_label.set_visible(False)

    def unbind(self) -> None:
        """Unbind category from row."""
        for binding in self.__bindings:
            binding.unbind()
        self.__bindings.clear()
        self.__category = None

    @Gtk.Template.Callback()
    def _activate(self, click: Gtk.GestureClick, _n_press: int, _x: float, _y: float) -> None:
        assert self.__category
        state = click.get_current_event_state()
        if state == Gdk.ModifierType.ALT_MASK:
            list_row = self.get_list_row()
            if list_row.is_expandable():
                list_row.set_expanded(not list_row.get_expanded())
        else:
            close_sidebar = state != Gdk.ModifierType.CONTROL_MASK
            if ConfigManager.get_default().pin_sidebar:
                self.__activated_func(self.__category, close_sidebar)
            else:
                # Without the idle callback the listview selection doesn't get set before the
                # sidebar closes (when auto-hiding is enabled)
                GLib.idle_add(self.__activated_func, self.__category, close_sidebar)

    def __add_binding(
        self,
        from_object: GObject.Object,
        from_property: str,
        to_object: GObject.Object,
        to_property: str,
        binding_flags: GObject.BindingFlags = GObject.BindingFlags.SYNC_CREATE,
    ) -> None:
        binding = from_object.bind_property(from_property, to_object, to_property, binding_flags)
        self.__bindings.append(binding)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import time
import unittest
import warnings

from iotas.note import Note

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_duplicate(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.excerpt = note.title
        note.favourite = False
        note.category = "category"

        dupe = note.duplicate()
        self.assertIsNotNone(dupe)
        self.assertEqual(dupe.content, "content")
        self.assertEqual(dupe.title, "Test Note Title")
        self.assertEqual(dupe.excerpt, "Test Note Title")
        self.assertEqual(dupe.favourite, False)
        self.assertEqual(dupe.category, "category")

    def test_flag_changed(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.saving = False
        self.assertFalse(note.dirty)
        note.flag_changed(update_last_modified=False)
        self.assertTrue(note.dirty)
        self.assertEqual(note.last_modified, 0)
        self.assertFalse(note.dirty_while_saving)

        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.saving = False
        self.assertFalse(note.dirty)
        note.flag_changed(update_last_modified=True)
        self.assertTrue(note.dirty)
        self.assertTrue(note.last_modified > 0)
        self.assertFalse(note.dirty_while_saving)

        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.saving = True
        note.flag_changed(update_last_modified=False)
        self.assertTrue(note.dirty)
        self.assertEqual(note.last_modified, 0)
        self.assertTrue(note.dirty_while_saving)

    def test_regenerate_excerpt(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        self.assertEqual(note.excerpt, "")
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "content")

        self.assertFalse(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "content")

        note.content = ""
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "")

        note.content = "# updated content"
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "updated content")

        note.content = "## Test Note Title\npost-title"
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "post-title")

        note.content = "- one\n+ two\n* three"
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "• one   • two   • three")

        note.content = "- [ ] one\n- [x] two"
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(note.excerpt, "☐ one   ☑ two")

        note.content = """long long long long long long long long long long long long long long
        long long long long long long long long long long long long long long long long long long
        long long long long long long long long long long long long long long long long long long
        long long long long long long long long long long long long long long long long long long
        long long long long long long long long long long long long long long long long long long
        long long long long long long long long long long long long long long long long long long
        """
        self.assertTrue(note.regenerate_excerpt())
        self.assertEqual(len(note.excerpt), note.EXCERPT_LENGTH)

    def test_repopulate_meta_from_sync_update(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.title_is_top_line = False

        modified = int(time.time())
        meta = {
            "title": "New Title",
            "category": "category",
            "modified": modified,
            "etag": "etag",
            "id": "remoteid",
        }
        note.repopulate_meta_from_sync_update(meta)
        self.assertEqual(note.title, "New Title")
        self.assertEqual(note.category, "category")
        self.assertEqual(note.last_modified, modified)
        self.assertEqual(note.etag, "etag")
        self.assertEqual(note.remote_id, "remoteid")

        note.title = "Reset Title"
        note.title_is_top_line = True
        note.repopulate_meta_from_sync_update(meta)
        self.assertEqual(note.title, "Reset Title")
        self.assertEqual(note.server_sanitised_title, "New Title")

    def test_update_from_remote_sync(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        modified = int(time.time())
        remote_update = {
            "title": "New Title",
            "content": "new content",
            "category": "category",
            "modified": modified,
            "etag": "etag",
            "readonly": True,
            "favorite": True,
        }

        note.update_from_remote_sync(remote_update)
        self.assertEqual(note.title, "New Title")
        self.assertEqual(note.content, "new content")
        self.assertEqual(note.excerpt, "new content")
        self.assertEqual(note.category, "category")
        self.assertEqual(note.last_modified, modified)
        self.assertEqual(note.etag, "etag")
        self.assertTrue(note.favourite)
        self.assertTrue(note.read_only)

        remote_update["readonly"] = False
        remote_update["favorite"] = False
        note.update_from_remote_sync(remote_update)
        self.assertFalse(note.favourite)
        self.assertFalse(note.read_only)

    def test_update_title_from_top_line(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.title_is_top_line = True
        note.update_title_from_top_line()
        self.assertEqual(note.title, "content")

        note.content = "content"
        note.title = "Test Note Title"
        note.title_is_top_line = False
        note.update_title_from_top_line()
        self.assertEqual(note.title, "Test Note Title")

    def test_identical_excepting_sync_meta(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.category = "category"
        note.favourite = True
        note.locally_deleted = False
        note.remote_id = 7
        note.last_modified = int(time.time())
        note.etag = "etag"
        note.dirty = True

        second = Note(new_note=True)
        second.content = "content"
        second.title = "Test Note Title"
        second.category = "category"
        second.favourite = True
        second.locally_deleted = False
        second.remote_id = 7
        second.last_modified = 0
        second.etag = "otheretag"
        second.dirty = False

        self.assertTrue(note.identical_excepting_sync_meta(second))
        second.content = "changed"
        self.assertFalse(note.identical_excepting_sync_meta(second))

    def test_identical_to(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.category = "category"
        note.favourite = True
        note.locally_deleted = False
        note.remote_id = 7
        note.last_modified = int(time.time())
        note.etag = "etag"
        note.dirty = True

        second = Note(new_note=True)
        second.content = "content"
        second.title = "Test Note Title"
        second.category = "category"
        second.favourite = True
        second.locally_deleted = False
        second.remote_id = 7
        second.last_modified = note.last_modified
        second.etag = "etag"
        second.dirty = True

        self.assertTrue(note.identical_to(second))
        second.dirty = False
        self.assertFalse(note.identical_to(second))

    def test_etag_is_recent(self) -> None:
        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"

        self.assertFalse(note.etag_is_recent(""))
        note.etag = "etag"

        self.assertTrue(note.etag_is_recent("etag"))
        self.assertFalse(note.etag_is_recent("missing"))
        self.assertFalse(note.etag_is_recent(""))

        note.etag = "second"
        self.assertTrue(note.etag_is_recent("second"))
        self.assertTrue(note.etag_is_recent("etag"))

        note = Note(new_note=True)
        note.content = "content"
        note.title = "Test Note Title"
        note.etag = ""

        self.assertFalse(note.etag_is_recent(""))

    def test_from_backup(self) -> None:
        modified = int(time.time())
        details = {
            "Title": "New Title",
            "Category": "category",
            "Favourite": True,
            "LastModified": modified,
            "LocallyDeleted": False,
            "Dirty": False,
            "ETag": "etag",
            "RemoteId": 7,
        }
        note = Note.from_backup(details, "content", include_sync_attributes=False)
        self.assertIsNotNone(note)
        assert note  # mypy
        self.assertEqual(note.title, "New Title")
        self.assertEqual(note.content, "content")
        self.assertEqual(note.excerpt, "content")
        self.assertEqual(note.category, "category")
        self.assertEqual(note.last_modified, modified)
        self.assertEqual(note.favourite, True)
        self.assertEqual(note.read_only, False)

        note = Note.from_backup(details, "content", include_sync_attributes=True)
        self.assertIsNotNone(note)
        assert note  # mypy
        self.assertEqual(note.etag, "etag")
        self.assertEqual(note.remote_id, 7)

    def test_from_nextcloud(self) -> None:
        modified = int(time.time())
        remote = {
            "title": "New Title",
            "content": "new content",
            "category": "category",
            "modified": modified,
            "etag": "etag",
            "readonly": True,
            "favorite": True,
            "id": 7,
        }
        note = Note.from_nextcloud(remote)

        self.assertIsNotNone(note)
        self.assertEqual(note.title, "New Title")
        self.assertEqual(note.content, "new content")
        self.assertEqual(note.excerpt, "new content")
        self.assertEqual(note.category, "category")
        self.assertEqual(note.last_modified, modified)
        self.assertEqual(note.etag, "etag")
        self.assertEqual(note.read_only, True)
        self.assertEqual(note.favourite, True)
        self.assertEqual(note.remote_id, 7)

    def test_new_and_empty(self) -> None:
        note = Note(new_note=True)
        note.content = ""
        self.assertTrue(note.new_and_empty)

        note.content = "content"
        self.assertFalse(note.new_and_empty)

        note.content = ""
        note.id = 7
        self.assertFalse(note.new_and_empty)

    def test_has_id(self) -> None:
        note = Note(new_note=True)
        self.assertFalse(note.has_id)
        note.id = 13
        self.assertTrue(note.has_id)

    def test_has_remote_id(self) -> None:
        note = Note(new_note=True)
        self.assertFalse(note.has_remote_id)
        note.remote_id = 13
        self.assertTrue(note.has_remote_id)

    def test_content_loaded(self) -> None:
        note = Note(new_note=True)
        self.assertFalse(note.content_loaded)
        note.content = ""
        self.assertTrue(note.content_loaded)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib, GObject

from typing import Callable, Optional
import unittest
import warnings

from iotas.attachment import Attachment
from iotas.batch_exporter import BatchExporter
from iotas.default_note_exporter import DefaultNoteExporter
from iotas.note import Note
from iotas.note_exporter import NoteExporter

warnings.filterwarnings("ignore", "version")


class _MockNoteExporter(NoteExporter):

    __gsignals__ = {
        # str: single export out path
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        # str: reason
        "failed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    note: Note
    out_format: str
    file_extension: str
    supporting_tex: bool
    allow_missing_images: bool
    user_location: Optional[str]

    export_calls: int = 0

    def export(
        self,
        note: Note,
        out_format: str,
        file_extension: str,
        supporting_tex: bool,
        allow_missing_images: bool = False,
        user_location: Optional[str] = None,
    ) -> None:
        self.note = note
        self.out_format = out_format
        self.file_extension = file_extension
        self.supporting_tex = supporting_tex
        self.allow_missing_images = allow_missing_images
        self.user_location = user_location
        self.export_calls += 1
        self.emit("finished", "")


class Test(unittest.TestCase):

    exporter: BatchExporter
    note_exporter: _MockNoteExporter

    load_content_calls: int
    download_attachment_calls: int

    def test_check_and_transfer_missing_attachments(self) -> None:
        exporter = self.__reset()
        note1 = Note(new_note=True)
        note1.id = 1
        note1.content = "![](image.png)"
        note1.title = "Test Note Title"

        note2 = Note(new_note=True)
        note2.id = 2
        note2.content = "simple"
        note2.title = "Test Note Second"

        exporter.set_notes([note1, note2])

        self.__call_and_wait(exporter.check_and_transfer_missing_attachments)
        self.assertEqual(self.load_content_calls, 0)
        self.assertEqual(self.download_attachment_calls, 1)
        self.assertEqual(self.note_exporter.export_calls, 0)

    def test_check_and_transfer_missing_attachments_no_attachments(self) -> None:
        exporter = self.__reset()
        note1 = Note(new_note=True)
        note1.id = 1
        note1.content = "plain"
        note1.title = "Test Note Title"

        exporter.set_notes([note1])

        self.__call_and_wait(exporter.check_and_transfer_missing_attachments)
        self.assertEqual(self.load_content_calls, 0)
        self.assertEqual(self.download_attachment_calls, 0)
        self.assertEqual(self.note_exporter.export_calls, 0)

    def test_check_and_transfer_missing_attachments_content_load(self) -> None:
        exporter = self.__reset()
        note1 = Note(new_note=True)
        note1.id = 1
        note1.content = None
        note1.title = "Test Note Title"
        self.assertFalse(note1.content_loaded)

        exporter.set_notes([note1])

        self.__call_and_wait(exporter.check_and_transfer_missing_attachments)
        self.assertEqual(self.load_content_calls, 1)
        self.assertEqual(self.download_attachment_calls, 0)
        self.assertEqual(self.note_exporter.export_calls, 0)
        self.assertTrue(note1.content_loaded)

    def test_start_export(self) -> None:
        exporter = self.__reset()
        note1 = Note(new_note=True)
        note1.id = 1
        note1.content = "plain"
        note1.title = "Test Note Title"

        note2 = Note(new_note=True)
        note2.id = 2
        note2.content = "simple"
        note2.title = "Test Note Second"

        exporter.set_notes([note1, note2])
        exporter.out_format = "md"
        exporter.file_extension = "md"
        exporter.supporting_tex = False
        exporter.location = "a-path"

        self.__call_and_wait(exporter.start_export)
        self.assertEqual(self.load_content_calls, 0)
        self.assertEqual(self.download_attachment_calls, 0)
        self.assertEqual(self.note_exporter.export_calls, 2)

        self.assertEqual(self.note_exporter.out_format, "md")
        self.assertEqual(self.note_exporter.file_extension, "md")
        self.assertEqual(self.note_exporter.supporting_tex, False)
        self.assertTrue(
            self.note_exporter.user_location
            in ["a-path/Test Note Title.md", "a-path/Test Note Second.md"]
        )
        self.assertEqual(self.note_exporter.allow_missing_images, False)

    def __reset(self) -> BatchExporter:
        self.note_exporter = _MockNoteExporter()
        self.exporter = BatchExporter(
            self.note_exporter, self.__mock_download_attachment, self.__mock_load_content
        )
        self.exporter.connect(
            "finished-attachment-detection-transfer-run", self.__on_finished_attachment_run
        )
        self.exporter.connect("batch-finished", self.__on_batch_finished)
        self.exporter.connect("export-failed", self.__on_export_failed)
        self.load_content_calls = 0
        self.download_attachment_calls = 0
        return self.exporter

    def __call_and_wait(self, call: Callable, *args, **kwargs) -> None:
        context = GLib.MainContext.default()
        self.success = False
        self.awaiting_signal = True
        call(*args, **kwargs)
        while self.awaiting_signal:
            context.iteration(False)

    def __on_finished_attachment_run(self, _obj: GObject.Object, count: int) -> None:
        self.awaiting_signal = False
        self.success = count == 0
        print("success?")

    def __on_batch_finished(self, *args) -> None:
        self.awaiting_signal = False
        self.success = True
        print("success?")

    def __on_export_failed(self, _obj: DefaultNoteExporter, error) -> None:
        print(f"Failed: {error}")
        self.awaiting_signal = False

    def __mock_download_attachment(self, attachment: Attachment) -> bool:
        self.download_attachment_calls += 1
        return True

    def __mock_load_content(self, note: Note) -> None:
        self.load_content_calls += 1
        note.content = "mock"



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gdk, Gio, Gtk

from typing import Callable, Optional

from iotas.category import Category, CategorySpecialPurpose
from iotas.category_manager import CategoryManager
from iotas.index_menu_button import IndexMenuButton
from iotas.sidebar_row import SidebarRow
from iotas.sync_manager import SyncManager
from iotas.theme_selector import ThemeSelector


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/sidebar.ui")
class Sidebar(Adw.Bin):
    __gtype_name__ = "Sidebar"

    _listview: Gtk.ListView = Gtk.Template.Child()
    _close_button: Gtk.Button = Gtk.Template.Child()
    _menu_button: IndexMenuButton = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self.__init_listview()

    def setup(
        self,
        sync_manager: SyncManager,
        category_manager: CategoryManager,
        category_activated_func: Callable[[Category, bool], None],
    ) -> None:
        """Perform initial setup.

        :param SyncManager sync_manager: Remote sync manager
        :param CategoryManager: Category manager
        :param Callable[[Category], None] category_activated_func: Function to call upon category
            activation.
        """
        self.__sync_manager = sync_manager
        self.__category_manager = category_manager
        self.__category_activated_func = category_activated_func
        self.__filter_model.set_model(self.__category_manager.list_model)
        self.__category_manager.connect("initial-load-complete", lambda _o: self.__populate())
        self._menu_button.get_popover().add_child(ThemeSelector(), "theme")

    def select_and_fetch_category(
        self,
        category_name: Optional[str],
        purpose: Optional[CategorySpecialPurpose] = None,
    ) -> Optional[Category]:
        """Select and fetch the specified category.

        :param Optional[str]: Name of category
        :param Optional[CategorySpecialPurpose]: Alternatively, a category special purpose
        :return: The category, if found
        :rtype: Optional[Category]
        """
        category = None
        if purpose is not None:
            category, index = self.__category_manager.get_special_purpose_category(purpose)
        elif category_name is not None:
            category, index = self.__category_manager.get_normal_category(category_name)
        else:
            raise Exception("Incorrect usage")
        self.__selection_model.set_selected(index)
        return category

    def take_focus(self) -> None:
        """Grab keyboard focus."""
        self._listview.grab_focus()
        # Ensure one of the rows has focus, but don't force it to the top row
        if self._listview.has_focus():
            for row in self._listview:
                row.grab_focus()
                break

    def has_focus_within(self) -> bool:
        """Fetch whether the sidebar or any of its rows have focus.

        :return: Whether focused
        :rtype: bool
        """
        # There's considerably more complexity here than expected due to Gtk.Widget.is_ancestor
        # not behaving as expected.
        if self._listview.has_focus():
            return True
        else:
            for row in self._listview:
                sidebar_row = row.get_first_child()
                if sidebar_row.has_focus():
                    return True
        return False

    def show_buttons(self, close_visible: bool, menu_visible: bool) -> None:
        """Set visibility on buttons.

        :param bool close_visible: Close button visibility
        :param bool menu_visible: Menu button visibility
        """
        self._close_button.set_visible(close_visible)
        self._menu_button.set_visible(menu_visible)

    def __init_listview(self) -> None:
        self.__factory = Gtk.SignalListItemFactory()
        self.__factory.connect("setup", self.__setup_listitem)
        self.__factory.connect("bind", self.__bind_listitem)
        self.__factory.connect("unbind", self.__unbind_listitem)
        self._listview.set_factory(self.__factory)

        self.__filter_model = Gtk.FilterListModel()
        self.__custom_filter = Gtk.CustomFilter()
        self.__filter_model.set_filter(self.__custom_filter)

        def filter_root_categories(category: Category) -> bool:
            return "/" not in category.name

        self.__custom_filter.set_filter_func(filter_root_categories)

    def __populate(self) -> None:
        self.__tree_model = Gtk.TreeListModel.new(
            self.__filter_model, False, False, self.__create_subcategory_models
        )
        self.__selection_model = Gtk.SingleSelection.new(self.__tree_model)
        self._listview.set_model(self.__selection_model)

        self._listview.connect("activate", self.__on_listview_row_activated)

        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__on_key_pressed)
        self._listview.add_controller(controller)

    def __create_subcategory_models(self, category: Category) -> Optional[Gio.ListModel]:
        if not category.is_sub_category:

            def filter_toplevel_parent(category: Category, parent_name: str) -> bool:
                return parent_name != category.name and category.name.startswith(parent_name + "/")

            filter = Gtk.CustomFilter.new(filter_toplevel_parent, category.name)
            model = Gtk.FilterListModel.new(self.__category_manager.list_model)
            model.set_filter(filter)
            if model.get_n_items() > 0:
                model.connect("notify::n-items", self.__subcategory_store_items_changed, category)
                return model
        return None

    def __setup_listitem(self, _factory: Gtk.SignalListItemFactory, listitem: Gtk.ListItem) -> None:
        row = SidebarRow(self.__category_activated_func, self.__category_manager)
        listitem.set_child(row)

    def __bind_listitem(self, _factory: Gtk.SignalListItemFactory, listitem: Gtk.ListItem) -> None:
        expander = listitem.get_child()
        tree_list_row = listitem.get_item()
        expander.set_list_row(tree_list_row)
        category = tree_list_row.get_item()
        expander.bind(category)

    def __unbind_listitem(
        self, _factory: Gtk.SignalListItemFactory, listitem: Gtk.ListItem
    ) -> None:
        row = listitem.get_child()
        row.unbind()

    def __subcategory_store_items_changed(
        self, model: Gtk.FilterListModel, _n_items: int, category: Category
    ) -> None:
        if model.get_n_items() == 0:
            self.__category_manager.recreate_category_for_expandable_change(category)

    def __expand_selected_row(self, expanded: bool) -> None:
        selected = self.__selection_model.get_selected()
        selected_row = self.__selection_model.get_item(selected)
        if selected_row.is_expandable() and selected_row.get_expanded() is not expanded:
            selected_row.set_expanded(expanded)

    def __on_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ):
        if keyval in (Gdk.KEY_Right, Gdk.KEY_KP_Right):
            self.__expand_selected_row(True)
            return Gdk.EVENT_STOP
        elif keyval in (Gdk.KEY_Left, Gdk.KEY_KP_Left):
            self.__expand_selected_row(False)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def __on_listview_row_activated(self, listview: Gtk.ListView, position: int) -> None:
        model = listview.get_model()
        model.select_item(position, True)
        row = model.get_item(position)
        category = row.get_item()
        close_sidebar = True
        self.__category_activated_func(category, close_sidebar)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("Secret", "1")
from gi.repository import Gdk, GObject, GLib, Gtk, Secret

import datetime
from email.utils import parsedate_to_datetime
from enum import IntEnum, auto
from json.decoder import JSONDecodeError
import logging
import os
import requests
from threading import Thread
from typing import NamedTuple, Optional
from ssl import SSLCertVerificationError

import iotas.const as const

from iotas.attachment import Attachment
from iotas.nextcloud_sync_worker_configuration import NextcloudSyncWorkerConfiguration
from iotas.note import Note
from iotas.sync_result import SyncResult, ContentPushSyncResult


class LoginRequestResult(IntEnum):

    SUCCESS = auto()
    FAILURE_GENERIC = auto()
    FAILURE_SELF_SIGNED_SSL = auto()
    FAILURE_CERTIFICATE = auto()


class AttachmentDownloadingResult(NamedTuple):

    data: Optional[bytes]
    status_code: int


class NextcloudSyncWorker(GObject.Object):
    __gsignals__ = {
        "ready": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "api-check-finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "settings-fetched": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "secret-service-failure": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "missing-password": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "notes-capability-missing": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    ENDPOINT_V_1_0 = "/index.php/apps/notes/api/v1/"
    ENDPOINT_V_1_4 = "/index.php/apps/notes/api/v1.4/"
    EXPECTED_API_VERSION = "1.4"

    LOGIN_ENDPOINT = "/index.php/login/v2"
    CAPABILITIES_ENDPOINT = "/ocs/v2.php/cloud/capabilities"

    SECRET_ATTRIBUTES = {"service": "nextcloud"}
    SECRET_SCHEMA = Secret.Schema.new(
        "org.gnome.World.Iotas",
        Secret.SchemaFlags.NONE,
        {"service": Secret.SchemaAttributeType.STRING},
    )

    USER_AGENT = f"Iotas/{const.VERSION}"

    CA_CHAIN_FILE = os.path.join(GLib.get_user_data_dir(), "iotas", "server_ssl_ca_chain.pem")

    # From openssl/x509_vfy.h
    OPENSSL_SELF_SIGNED_CERTIFICATE_ERROR_CODES = (18, 19)

    def __init__(self, config_manager: NextcloudSyncWorkerConfiguration) -> None:
        super().__init__()
        self.__config_manager = config_manager
        self.__app_password: Optional[str] = None
        self.__secret_service_failed = False
        self.authenticated = False
        self.__timeout = 10
        self.__api_version = ""

        self.__server = self.__config_manager.nextcloud_endpoint
        self.__username = self.__config_manager.nextcloud_username
        self.__prune_threshold = self.__config_manager.nextcloud_prune_threshold

        if os.path.exists(self.CA_CHAIN_FILE):
            logging.info("Using SSL CA chain file at {}".format(self.CA_CHAIN_FILE))
            self.verify: str | bool = self.CA_CHAIN_FILE
        else:
            self.verify = True

        self.__headers = {
            "User-Agent": self.USER_AGENT,
            "Cache-Control": "no-cache",
        }

    def init_auth(self) -> None:
        """Initialise Nextcloud authentication details from the Secret Service."""
        thread = Thread(target=self.__init_secrets)
        thread.daemon = True
        thread.start()

    def start_login(
        self, server_uri: str, window: Gtk.Window
    ) -> tuple[LoginRequestResult, Optional[str], Optional[str]]:
        """Initiate a new login.

        :param str server_uri: Server instance URI
        :param Gtk.Window window: Main window
        :return: tuple with LoginRequestResult before endpoint URI and auth token (or Nones on
            failures)
        :rtype: tuple[LoginRequestResult, Optional[str], Optional[str]]
        """
        uri = "{}{}".format(server_uri, self.LOGIN_ENDPOINT)
        try:
            res = requests.post(
                uri,
                headers=self.__headers,
                verify=self.verify,
                timeout=self.__timeout,
            )
        except requests.exceptions.SSLError as e:
            if self.exception_is_for_self_signed_certificate(e):
                logging.warning("Login found self-signed certificate: {} for {}".format(e, uri))
                return (LoginRequestResult.FAILURE_SELF_SIGNED_SSL, None, None)
            elif self.get_ssl_verification_error(e) is not None:
                logging.warning(
                    "Login POST encountered SSL certificate problem: {} for {}".format(e, uri)
                )
                return (LoginRequestResult.FAILURE_CERTIFICATE, None, None)
            else:
                logging.warning("Login POST request error: {} for {}".format(e, uri))
                return (LoginRequestResult.FAILURE_GENERIC, None, None)
        except Exception as e:
            logging.warning("Login POST request error: {} for {}".format(e, uri))
            return (LoginRequestResult.FAILURE_GENERIC, None, None)

        if res is None:
            logging.warning("Login POST response is null")
            return (LoginRequestResult.FAILURE_GENERIC, None, None)

        if res.status_code != 200:
            logging.warning(
                "Login POST request failed with status code: {}".format(res.status_code)
            )
            return (LoginRequestResult.FAILURE_GENERIC, None, None)

        try:
            response = res.json()
        except requests.exceptions.JSONDecodeError as e:
            logging.warning("Failed to decode login request to JSON: {} for {}".format(e, res.text))
            return (LoginRequestResult.FAILURE_GENERIC, None, None)

        # TODO change this to not need the window passed in
        def open_url_in_browser(window: Gtk.Window, response: dict) -> None:
            Gtk.show_uri(window, response["login"], Gdk.CURRENT_TIME)

        GLib.idle_add(open_url_in_browser, window, response)
        return (LoginRequestResult.SUCCESS, response["poll"]["endpoint"], response["poll"]["token"])

    def check_endpoint_for_auth_token(self, endpoint: str, token: str) -> tuple[bool, bool]:
        """Poll endpoint for authorisation success.

        :param str endpoint: Endpoint URI
        :param str token: Request token
        :return: Whether authenticated and whether auth. storage succeeded
        :rtype: tuple[bool, bool]
        """
        data = {"token": token}
        try:
            res = requests.post(
                endpoint,
                headers=self.__headers,
                data=data,
                verify=self.verify,
                timeout=self.__timeout,
            )
        except Exception:
            return (False, False)
        if res.status_code != 200:
            return (False, False)

        response = res.json()

        stored = self.__store_password(response["loginName"], response["appPassword"])

        if stored:
            self.__authenticated_and_password_stored(
                response["server"], response["loginName"], response["appPassword"]
            )
        else:
            self.__handle_secret_service_failure(emit=False)

        return (True, stored)

    def sign_out(self) -> None:
        """Clear credentials." """
        logging.debug("Nextcloud signout")
        self.__clear_password()
        self.reset_prune_threshold()

    def check_capabilities(self) -> None:
        """Check server API version."""
        assert self.__username is not None
        assert self.__app_password is not None
        url = self.__server + "/" + self.CAPABILITIES_ENDPOINT
        headers = self.__headers.copy()
        headers["OCS-APIRequest"] = "true"
        headers["Accept"] = "application/json"
        try:
            res = requests.get(
                url,
                headers=headers,
                auth=(self.__username, self.__app_password),
                verify=self.verify,
                timeout=self.__timeout,
            )
        except Exception as e:
            logging.warning(f"GET request for capabilities failed: {e}")
            return

        response = self.__parse_get_response_to_json(res, "capabilities")
        if response is None:
            return
        try:
            capabilities = response["ocs"]["data"]["capabilities"]["notes"]
            api_version = capabilities["api_version"]
            app_version = capabilities["version"]
        except KeyError:
            logging.warning(
                "Couldn't access version elements in server capabilities response. "
                "Is the Nextcloud Notes app installed within Nextcloud?"
            )
            GLib.idle_add(self.emit, "notes-capability-missing")
            return

        logging.info(
            f"Running against Nextcloud Notes server app v{app_version} "
            f"with API versions: {', '.join(api_version)}"
        )
        if self.EXPECTED_API_VERSION not in api_version:
            logging.warning(
                f"Expected API version {self.EXPECTED_API_VERSION} not found in server API "
                f"versions ({', '.join(api_version)})"
            )
        self.__api_version = max(api_version)

        GLib.idle_add(self.emit, "api-check-finished")

    def fetch_settings(self) -> None:
        """Fetch settings from server."""
        res = self.__request_get("settings", {}, self.__timeout)
        if res is None:
            logging.warning("Couldn't fetch settings from server")
            return

        response = self.__parse_get_response_to_json(res, "settings")
        if response is None:
            return
        try:
            file_suffix = response["fileSuffix"]
        except KeyError:
            logging.warning("Couldn't access file suffix setting in server response")
            return

        file_suffix = file_suffix.strip()
        file_suffix = file_suffix.strip(".")
        GLib.idle_add(self.emit, "settings-fetched", file_suffix)

    def get_notes(self) -> SyncResult:
        """Fetch notes from server.

        :return: Sync result object
        :rtype: SyncResult
        """
        params = {}
        if self.__prune_threshold > 0:
            params["pruneBefore"] = self.__prune_threshold

        # Bump network timeout for initial imports, long breaks between syncs and timestamp resets
        two_weeks_ago = (datetime.datetime.now() - datetime.timedelta(weeks=2)).timestamp()
        if self.__prune_threshold > two_weeks_ago:
            timeout = self.__timeout
        else:
            # Two minutes, based on default 10 second timeout
            timeout = self.__timeout * 12
            reason = "Over two weeks since sync" if self.__prune_threshold else "Fetching all notes"
            logging.info(
                f"{reason}, using longer network timeout for potential longer sync ({timeout:.2f}s)"
            )

        res = self.__request_get("notes", params, timeout)
        ret = SyncResult.from_requests_response(res)
        if not ret.success:
            return ret
        if res and "Last-Modified" in res.headers:
            dt = parsedate_to_datetime(res.headers["Last-Modified"])
            self.__prune_threshold = int(dt.timestamp())

        return ret

    def update_note(self, note: Note) -> ContentPushSyncResult:
        """Update a note.

        :param Note note: Note to update
        :return: Result of operation
        :rtype: ContentPushSyncResult
        """
        headers = {"If-Match": '"' + note.etag + '"'}
        headers.update(self.__headers)
        path = f"notes/{note.remote_id}"
        sent_content = note.content
        data = {
            "title": note.title,
            "content": sent_content,
            "category": note.category,
            "favorite": 1 if note.favourite else 0,
        }
        res = self.__request_put(path, headers, data)
        return ContentPushSyncResult.from_requests_response(res, sent_content)

    def create_note(self, note: Note) -> ContentPushSyncResult:
        """Create a note.

        :param Note note: Note to create
        :return: Result of operation
        :rtype: ContentPushSyncResult
        """
        path = "notes"
        sent_content = note.content
        data = {"title": note.title, "content": sent_content, "category": note.category}
        res = self.__request_post(path, data)
        return ContentPushSyncResult.from_requests_response(res, sent_content)

    def delete_note(self, note: Note) -> SyncResult:
        """Delete a note.

        :param Note note: Note to delete
        :return: Result of operation
        :rtype: SyncResult
        """
        path = "notes/{}".format(note.remote_id)
        res = self.__request_delete(path)
        return SyncResult.from_requests_response(res)

    def get_attachment(self, attachment: Attachment) -> AttachmentDownloadingResult:
        """Fetch an attachment from the server.

        :param Attachment attachment: Attachment
        :return: Result object
        :rtype: AttachmentDownloadingResult
        """
        params = {}
        params["path"] = attachment.path

        res = self.__request_get(f"attachment/{attachment.note_remote_id}", params, self.__timeout)

        if res is None:
            return AttachmentDownloadingResult(None, -1)
        elif res.status_code != 200:
            return AttachmentDownloadingResult(None, res.status_code)
        else:
            return AttachmentDownloadingResult(res.content, res.status_code)

    def reset_prune_threshold(self) -> None:
        """Reset the sync threshold."""
        self.__prune_threshold = 0
        self.__config_manager.nextcloud_prune_threshold = self.__prune_threshold

    @GObject.Property(type=str)
    def server(self) -> str:
        return self.__server

    @server.setter
    def set_server(self, value: str) -> None:
        value = value.strip()
        value = GLib.uri_escape_string(value, None, True)
        value = value.strip("/")
        self.__server = value

    @GObject.Property(type=bool, default=False)
    def authenticated(self) -> bool:
        return self.__authenticated

    @authenticated.setter
    def set_authenticated(self, value: bool) -> None:
        self.__authenticated = value

    @GObject.Property(type=int)
    def prune_threshold(self) -> int:
        return self.__prune_threshold

    @GObject.Property(type=bool, default=False)
    def secret_service_failed(self) -> bool:
        return self.__secret_service_failed

    @GObject.Property(type=int)
    def timeout(self) -> int:
        return self.__timeout

    @timeout.setter
    def set_timeout(self, value: int) -> None:
        self.__timeout = value

    @GObject.Property(type=str)
    def api_version(self) -> str:
        return self.__api_version

    @api_version.setter
    def set_api_version(self, value: str) -> None:
        self.__api_version = value

    def __init_secrets(self) -> None:
        # Doing this sync fetch works around some type of libsecret flatpak issue (... so we may
        # as well use the result)
        self.__app_password = self.__fetch_password()
        if self.__app_password is None:
            self.emit("missing-password")
        else:
            self.__check_setup()
        logging.debug("Nextcloud sync setup: {}".format(self.__authenticated))

    def __check_setup(self) -> None:
        if (
            self.__server.strip() != ""
            and self.__username.strip() != ""
            and self.__app_password is not None
        ):
            self.authenticated = True
            self.emit("ready", False)

    def __authenticated_and_password_stored(
        self, endpoint: str, username: str, app_password: str
    ) -> None:
        self.__server = endpoint
        self.__username = username
        self.__app_password = app_password
        self.__config_manager.nextcloud_endpoint = self.__server
        self.__config_manager.nextcloud_username = self.__username

        self.reset_prune_threshold()
        self.authenticated = True
        GLib.idle_add(self.emit, "ready", True)
        logging.debug("Nextcloud login successful")

    def __request_get(self, path: str, params: dict, timeout: float) -> Optional[requests.Response]:
        assert self.__username is not None
        assert self.__app_password is not None
        url = self.__get_url(path)
        try:
            return requests.get(
                url,
                headers=self.__headers,
                auth=(self.__username, self.__app_password),
                params=params,
                verify=self.verify,
                timeout=timeout,
            )
        except Exception as e:
            logging.warning(f"GET request failed: {e}")
            return None

    def __request_put(self, path: str, headers: dict, data: dict) -> Optional[requests.Response]:
        assert self.__username is not None
        assert self.__app_password is not None
        url = self.__get_url(path)
        try:
            return requests.put(
                url,
                headers=headers,
                auth=(self.__username, self.__app_password),
                data=data,
                verify=self.verify,
                timeout=self.__timeout,
            )
        except Exception as e:
            logging.warning(f"PUT request failed: {e}")
            return None

    def __request_post(
        self, path: str, data: dict, files: Optional[dict] = None
    ) -> Optional[requests.Response]:
        assert self.__username is not None
        assert self.__app_password is not None
        url = self.__get_url(path)
        try:
            return requests.post(
                url,
                headers=self.__headers,
                auth=(self.__username, self.__app_password),
                data=data,
                files=files,
                verify=self.verify,
                timeout=self.__timeout,
            )
        except Exception as e:
            logging.warning(f"POST request failed: {e}")
            return None

    def __request_delete(self, path: str) -> Optional[requests.Response]:
        assert self.__username is not None
        assert self.__app_password is not None
        url = self.__get_url(path)
        try:
            return requests.delete(
                url,
                headers=self.__headers,
                auth=(self.__username, self.__app_password),
                verify=self.verify,
                timeout=self.__timeout,
            )
        except Exception as e:
            logging.warning(f"DELETE request failed: {e}")
            return None

    def __get_url(self, path: str) -> str:
        endpoint = self.ENDPOINT_V_1_4 if "attachment" in path else self.ENDPOINT_V_1_0
        return self.__server + "/" + endpoint + path

    def __store_password(self, login: str, password: str) -> bool:
        success = False
        try:
            item_id = "org.gnome.World.Iotas: {}".format(login)
            success = Secret.password_store_sync(
                self.SECRET_SCHEMA,
                self.SECRET_ATTRIBUTES,
                Secret.COLLECTION_DEFAULT,
                item_id,
                password,
                None,
            )
        except Exception as e:
            logging.error(f"Failed to store password: {e}")
        return success

    def __fetch_password(self) -> Optional[str]:
        pwd = None
        try:
            pwd = Secret.password_lookup_sync(self.SECRET_SCHEMA, self.SECRET_ATTRIBUTES, None)
        except GLib.Error as e:
            if "ServiceUnknown" in e.message:
                self.__handle_secret_service_failure(emit=True)
            else:
                logging.error(f"Failed to fetch password: {e}")
        return pwd

    def __handle_secret_service_failure(self, emit: bool):
        self.__secret_service_failed = True
        msg = (
            "Could not connect with the Secret Service. Without the service Nextcloud "
            "authentication details cannot be stored. Look into GNOME Keyring. Restart "
            "the app to try again."
        )
        logging.warning(msg)
        if emit:
            self.emit("secret-service-failure")

    def __clear_password(self) -> None:
        logging.debug("Clearing Nextcloud credentials")
        self.authenticated = False
        self.__app_password = None
        try:
            Secret.password_clear_sync(self.SECRET_SCHEMA, self.SECRET_ATTRIBUTES, None)
        except Exception as e:
            logging.warning(f"Failed to clear password: {e}")

    def __parse_get_response_to_json(
        self, res: Optional[requests.Response], purpose: str
    ) -> Optional[dict]:
        if res is None:
            logging.warning(f"Response for {purpose} is null")
            return None

        if res.status_code != 200:
            logging.warning(
                "Request for {} failed with status code: {}".format(purpose, res.status_code)
            )
            return None

        try:
            response = res.json()
        except JSONDecodeError as e:
            logging.warning(
                "Failed to decode {} request to JSON: {} for {}".format(purpose, e, res.text)
            )
            return None

        return response

    @staticmethod
    def exception_is_for_self_signed_certificate(e: requests.exceptions.SSLError) -> bool:
        ssl_verification_error = NextcloudSyncWorker.get_ssl_verification_error(e)
        if ssl_verification_error is None:
            return False

        return (
            ssl_verification_error.verify_code
            in NextcloudSyncWorker.OPENSSL_SELF_SIGNED_CERTIFICATE_ERROR_CODES
        )

    @staticmethod
    def get_ssl_verification_error(
        e: requests.exceptions.SSLError,
    ) -> Optional[SSLCertVerificationError]:
        # Get the requests SSLError
        if type(e.args) is not tuple or len(e.args) != 1:
            return None
        max_retry_error = e.args[0]
        ssl_error = max_retry_error.reason

        # Get the ssl.SSLCertVerificationError
        if type(ssl_error.args) is not tuple or len(ssl_error.args) != 1:
            return None
        if not isinstance(ssl_error.args[0], SSLCertVerificationError):
            return None
        return ssl_error.args[0]



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib, GObject

from threading import Thread
from typing import Callable

import markdown_it
from markdown_it.tree import SyntaxTreeNode

from iotas.note import Note


class OutlineHeading(GObject.Object):
    """Heading details for outline."""

    def __init__(self, text: str, line: int, level: int) -> None:
        super().__init__()
        self.text = text
        self.line = line
        self.level = level
        self.indent_level = 0


class OutlineGenerator(GObject.Object):
    """Generate outline from note headings."""

    def generate(self, note: Note, callback: Callable) -> None:
        """Generator outline for note.

        :param Note note: Note to render
        :param Callable callback: Method to call on completion
        """

        def thread_do():
            md = markdown_it.MarkdownIt("gfm-like")
            parser_tokens = md.parse(note.content)
            node = SyntaxTreeNode(parser_tokens)
            headings = []
            levels = []

            def add_children(in_node):
                if not in_node.is_root and in_node.type == "heading":
                    level = len(in_node.markup)
                    content = ""
                    for child in in_node.children:
                        content += child.content
                    headings.append(OutlineHeading(content, in_node.map[0], level))
                    levels.append(level)
                else:
                    for child in in_node.children:
                        add_children(child)

            add_children(node)

            # Squash the indentation levels, eg. a note with only h1 and h5 headings will show
            # as first and second level indentation
            indentation = {}
            i = 0
            for level in sorted(set(levels)):
                indentation[level] = i
                i += 1

            for heading in headings:
                heading.indent_level = indentation[heading.level]
            GLib.idle_add(callback, headings)

        thread = Thread(target=thread_do)
        thread.daemon = True
        thread.start()
